#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to visualize NER Text Dataset.

Methods
-------
    render(items: List[LabeledTextItem], options: Dict = None) -> str
        Renders NER dataset to Html format.

Examples
--------
>>> record1 = LabeledTextItem("London is the capital of the United Kingdom", [NERItem('city', 0, 6), NERItem("country", 29, 14)])
>>> record2 = LabeledTextItem("Houston area contractor seeking a Sheet Metal Superintendent.", [NERItem("city", 0, 6)])
>>> result = render(items = [record1, record2], options={"default_color":"#DDEECC", "colors": {"city":"#DDEECC", "country":"#FFAAAA"}})
>>> display(HTML(result))
"""

import logging
from dataclasses import asdict, dataclass
from string import Template
from typing import Dict, List, Optional

import pandas as pd
from ads.data_labeling.constants import AnnotationType
from ads.data_labeling.ner import NERItem
from cerberus import Validator


logger = logging.getLogger(__name__)

HTML_OPTIONS_SCHEMA = {
    "default_color": {"nullable": True, "required": False, "type": "string"},
    "colors": {
        "nullable": True,
        "required": False,
        "type": "dict",
    },
}

DEFAULT_COLOR = "#cedddd"
LEN_OF_SPACY_ITEM = 3


@dataclass
class LabeledTextItem:
    """Data class representing NER Item.

    Attributes
    ----------
    txt: str
        The labeled sentence.
    ents: List[NERItem]
        The list of entities.
    """

    txt: str
    ents: List[NERItem]

    def _validate(self):
        """Validates the instance.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If txt is empty.
        WrongEntityFormat
            If the list of entities has a wrong format.
        AssertionError
            In case of overlapped entities.
        """
        if not self.txt:
            raise ValueError(
                "The parameter `txt` is required and must not be an empty string."
            )

        if not isinstance(self.ents, List):
            raise ValueError(
                "Invalid format for the entities. The entities must be a List[NERItem]."
            )

        for entity in self.ents:
            if entity.offset + entity.length > len(self.txt):
                raise ValueError(
                    f"At least one of the entities (start index: {entity.length}, offset: {entity.offset}) "
                    f"exceeds the length of the text ({len(self.txt)})."
                )

        self.ents.sort(key=lambda x: x.offset)

        for i in range(len(self.ents) - 1):
            if self.ents[i].offset + self.ents[i].length >= self.ents[i + 1].offset:
                raise AssertionError(
                    "The entity data contains overlapping tokens. The first token has a start index"
                    f" of {self.ents[i].length}, and an offset of {self.ents[i].offset}. The second token has a start "
                    f"index of {self.ents[i + 1].length}, and an offset of {self.ents[i + 1].offset}. "
                )

    def __post_init__(self):
        self._validate()


@dataclass
class RenderOptions:
    """Data class representing render options.

    Attributes
    ----------
    default_color: str
        The specified default color.
    colors: Optional[dict]
        The multiple specified colors.
    """

    default_color: str
    colors: Optional[dict]

    @staticmethod
    def _validate(options: dict) -> None:
        """Validate whether the options passed in fits the defined schema.

        Parameters
        ----------
        options: dict
            The multiple specified colors.

        Returns
        -------
        None
            Nothing.
        """
        if not options:
            return None

        validator = Validator(HTML_OPTIONS_SCHEMA)
        valid = validator.validate(options)
        if not valid:
            raise ValueError(validator.errors)

    @classmethod
    def from_dict(cls, options: dict) -> "RenderOptions":
        """Constructs an instance of RenderOptions from a dictionary.

        Parameters
        ----------
        options: dict
            Render options in dictionary format.

        Returns
        -------
        RenderOptions
            The instance of RenderOptions.
        """
        if not options:
            return cls(default_color=DEFAULT_COLOR, colors={})

        RenderOptions._validate(options)

        return cls(
            options.get("default_color", DEFAULT_COLOR), options.get("colors", {}) or {}
        )

    def to_dict(self):
        """Converts RenderOptions instance to dictionary format.

        Returns
        -------
        dict
            The render options in dictionary format.
        """
        return asdict(self)

    def __repr__(self) -> str:
        return self.to_dict()


class TextLabeledDataFormatter:
    """The TextLabeledDataFormatter class to render NER items into Html format."""

    _ITEM_TEMPLATE = (
        '<span style="background-color: $color; padding: 5px; margin: 0px 5px; border-radius: 5px;">'
        '<span style="margin-right: 5px;">$entity</span>'
        '<span style="text-transform: uppercase; font-weight: bold; font-size:0.8em;">$label</span>'
        "</span>"
    )
    _ROW_TEMPLATE = '<div key=$key style="margin-top:10px; line-height:2em">$row</div>'

    @staticmethod
    def render(items: List[LabeledTextItem], options: Dict = None) -> str:
        """Renders NER dataset to Html format.

        Parameters
        ----------
        items: List[LabeledTextItem]
            Items to render.
        options: Optional[dict]
            Render options.

        Returns
        -------
        str
            Html representation of rendered NER dataset.

        Raises
        ------
        ValueError
            If items not provided.
        TypeError
            If items provided in a wrong format.
        """
        if not items:
            raise ValueError("The parameter `items` is required.")

        if not isinstance(items, list) or not all(
            isinstance(x, LabeledTextItem) for x in items
        ):
            raise TypeError(
                "Wrong format for the items. Items should be a `List[LabeledTextItem]`."
            )

        render_options = RenderOptions.from_dict(options)
        item_template = Template(TextLabeledDataFormatter._ITEM_TEMPLATE)
        row_template = Template(TextLabeledDataFormatter._ROW_TEMPLATE)
        result = []

        for item_index, item in enumerate(items):
            current_index = 0
            accum = []
            for e in item.ents:
                start = e.offset
                end = e.offset + e.length
                label = e.label
                accum.append(item.txt[current_index:start])
                accum.append(
                    item_template.substitute(
                        {
                            "color": render_options.colors.get(
                                label, render_options.default_color
                            ),
                            "entity": item.txt[start:end],
                            "label": label,
                        }
                    )
                )
                current_index = end

            accum.append(item.txt[current_index : len(item.txt)])
            result.append(
                row_template.substitute({"key": str(item_index), "row": "".join(accum)})
            )

        return "".join(result)


def _df_to_ner_items(
    df: pd.DataFrame,
    content_column: str = "Content",
    annotations_column: str = "Annotations",
) -> List[LabeledTextItem]:
    """Converts pandas dataframe into a list of LabeledTextItem.

    Parameters
    ----------
    df: pd.DataFrame
        The Pandas dataframe to convert.
    content_column: Optional[str]
            The column name with the content data.
    annotations_column: Optional[str]
        The column name for the annotations list.

    Returns
    -------
    List[LabeledTextItem]
        The list of LabeledTextItem objects.

    Raises
    ------
    TypeError
        If input data not a pandas dataframe.
    ValueError
        If input data has wrong format.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("The parameter `df` must be a Pandas dataframe.")
    if content_column not in list(df.columns):
        raise ValueError(
            f"The parameter `df` is invalid. It must have a column named `{content_column}`."
        )
    if annotations_column not in list(df.columns):
        raise ValueError(
            f"The parameter `df` is invalid. It must have a column named `{annotations_column}`."
        )

    if df[content_column].isnull().values.any():
        logger.warning(
            "The parameter `df` includes records where the text content is not "
            "materialized. These records will be ignored. Use `materialize=True` "
            "to load the content."
        )

    result = []
    for item in df.T.to_dict().values():
        if item[annotations_column] and not isinstance(item[annotations_column], list):
            raise ValueError(
                f"The parameter `df` is invalid. The column {annotations_column} "
                "must be of type `List[NERItem]`."
            )

        if item[content_column]:
            if (
                isinstance(item[annotations_column][0], tuple)
                and len(item[annotations_column][0]) == LEN_OF_SPACY_ITEM
            ):
                ents = [NERItem.from_spacy(ent) for ent in item[annotations_column]]
            else:
                ents = item[annotations_column] or []
            result.append(
                LabeledTextItem(
                    txt=item[content_column],
                    ents=ents,
                )
            )

    return result


def render(items: List[LabeledTextItem], options: Dict = None) -> str:
    """Renders NER dataset to Html format.

    Parameters
    ----------
    items: List[LabeledTextItem]
        The list of NER items to render.
    options: dict, optional
        The options for rendering.

    Returns
    -------
    str
        Html string.

    Examples
    --------
    >>> record = LabeledTextItem("London is the capital of the United Kingdom", [NERItem('city', 0, 6), NERItem("country", 29, 14)])
    >>> result = render(items = [record], options={"default_color":"#DDEECC", "colors": {"city":"#DDEECC", "country":"#FFAAAA"}})
    >>> display(HTML(result))
    """
    return TextLabeledDataFormatter.render(items, options)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
credit card patterns refer to https://en.wikipedia.org/wiki/Payment_card_number#Issuer_identification_number_(IIN)
Active and frequent card information
American Express: 34, 37
Diners Club (US & Canada): 54,55
Discover Card: 6011, 622126 - 622925, 624000 - 626999, 628200 - 628899, 64, 65
Master Card: 2221-2720, 51–55
Visa: 4

"""


class card_identify:
    def identify_issue_network(self, card_number):

        """
        Returns the type of credit card based on its digits

        Parameters
        ----------
        card_number: String

        Returns
        -------
        String: A string corresponding to the kind of credit card.
        """
        try:
            first_six_digit = int(str(card_number)[:6])
            first_four_digit = int(str(card_number)[:4])
            first_two_digit = int(str(card_number)[:2])
            first_one_digit = int(str(card_number)[:1])

            if first_one_digit == 4:
                return "Visa"
            elif first_two_digit == 34 or first_two_digit == 37:
                return "Amex"
            elif first_two_digit == 54 or first_two_digit == 55:
                return "Diners Club"
            elif (
                first_two_digit == 64
                or first_two_digit == 65
                or first_four_digit == 6011
                or 622126 <= first_six_digit <= 622925
                or 624000 <= first_six_digit <= 626999
                or 628200 <= first_six_digit <= 628899
            ):
                return "Discover"
            elif 2221 <= first_four_digit <= 2720 or 51 <= first_two_digit <= 55:
                return "MasterCard"
            else:
                return "Unknown"
        except:
            return "Unknown"


if __name__ == "__main__":
    _visa = [
        "4532640527811543",
        "4556929308150929",
        "4539944650919740",
        "4485348152450846",
        "4556593717607190",
    ]
    _mastercard = [
        "5334180299390324",
        "5111466404826446",
        "5273114895302717",
        "5430972152222336",
        "5536426859893306",
    ]
    _amex = [
        "371025944923273",
        "374745112042294",
        "340984902710890",
        "375767928645325",
        "370720852891659",
    ]



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict

from ads.common import auth as auth_util
from ads.common import oci_client
from ads.config import JOB_RUN_COMPARTMENT_OCID, NB_SESSION_COMPARTMENT_OCID
from ads.data_labeling.interface.reader import Reader
from ads.data_labeling.metadata import Metadata
from ads.data_labeling.parser.export_metadata_parser import MetadataParser
from ads.data_labeling.reader.jsonl_reader import JsonlReader
from oci.exceptions import ServiceError


class EmptyMetadata(Exception):
    """Empty Metadata."""

    pass


class ReadDatasetError(Exception):
    def __init__(self, id: str):
        super().__init__(f"Error occurred in attempt to read dataset {id}.")


class DatasetNotFoundError(Exception):
    def __init__(self, id: str):
        super().__init__(f"{id} not found.")


class MetadataReader:
    """MetadataReader class which reads and extracts the labeled dataset metadata.

    Examples
    --------
    >>> from ads.data_labeling import MetadataReader
    >>> import oci
    >>> import os
    >>> from ads.common import auth as authutil
    >>> reader = MetadataReader.from_export_file("metadata_export_file_path",
    ...                                 auth=authutil.api_keys())
    >>> reader.read()
    """

    def __init__(self, reader: Reader):
        """Initiate a MetadataReader instance.

        Parameters
        ----------
        reader: Reader
            Reader instance which reads and extracts the labeled dataset metadata.
        """
        self._reader = reader

    @classmethod
    def from_export_file(cls, path: str, auth: Dict = None) -> "MetadataReader":
        """Contructs a MetadataReader instance.

        Parameters
        ----------
        path: str
            metadata file path, can be either local or object storage path.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        MetadataReader
            The MetadataReader instance whose reader is a ExportMetadataReader instance.
        """
        return cls(ExportMetadataReader(path=path, auth=auth))

    @classmethod
    def from_DLS(
        cls, dataset_id: str, compartment_id: str = None, auth: dict = None
    ) -> "MetadataReader":
        """Contructs a MetadataReader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        compartment_id: (str, optional). Default None
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        MetadataReader
            The MetadataReader instance whose reader is a DLSMetadataReader instance.
        """
        if compartment_id is None:
            compartment_id = NB_SESSION_COMPARTMENT_OCID or JOB_RUN_COMPARTMENT_OCID

        if not compartment_id:
            raise ValueError("The `compartment_id` must be provided.")

        auth = auth or auth_util.default_signer()

        return cls(
            DLSMetadataReader(
                dataset_id=dataset_id, compartment_id=compartment_id, auth=auth
            )
        )

    def read(self) -> Metadata:
        """Reads the content from the metadata file.

        Returns
        -------
        Metadata
            The metadata of the labeled dataset.
        """
        return self._reader.read()


class DLSMetadataReader(Reader):
    """DLSMetadataReader class which reads the metadata jsonl file from the cloud."""

    def __init__(self, dataset_id: str, compartment_id: str, auth: dict):
        """Initializes the DLS metadata reader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        compartment_id: str
            The compartment OCID of the dataset.
        auth: dict
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Raises
        ------
            ValueError: When dataset_id is empty or not a string.
            TypeError: When dataset_id not a string.
        """
        if not dataset_id:
            raise ValueError("The dataset OCID must be specified.")

        if not isinstance(dataset_id, str):
            raise TypeError("The dataset_id must be a string.")
        self.dataset_id = dataset_id
        self.compartment_id = compartment_id

        self.dls_dp_client = oci_client.OCIClientFactory(**auth).data_labeling_dp

    def read(self) -> Metadata:
        """Reads the content from the metadata file.

        Returns
        -------
        Metadata
            The metadata of the labeled dataset.

        Raises
        ------
        DatasetNotFoundError
            If dataset not found.
        ReadDatasetError
            If any error occured in attempt to read dataset.
        """
        try:
            dataset_response = self.dls_dp_client.get_dataset(self.dataset_id)
        except ServiceError as service_err:
            if service_err.status == 404:
                raise DatasetNotFoundError(self.dataset_id)
            raise ReadDatasetError(self.dataset_id)
        return Metadata.from_dls_dataset(dataset_response.data)


class ExportMetadataReader(JsonlReader):
    """ExportMetadataReader class which reads the metadata jsonl file from local/object
    storage path."""

    def read(self) -> Metadata:
        """Reads the content from the metadata file.

        Returns
        -------
        Metadata
            The metadata of the labeled dataset.
        """

        try:
            return MetadataParser.parse(next(super().read()))
        except StopIteration:
            raise EmptyMetadata(
                "The dataset metadata file is invalid. It appears to be empty. "
                "Use the `DataLabeling.export()` method to create a new dataset metadata file."
            )
        except Exception as e:
            raise e



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from .abstract_operator_spec import V0Spec, DistributedV1Spec
import os
from ads.opctl.utils import suppress_traceback
import yaml


class OperatorSpecFactory:
    """
    Factory class for creating provider instance.
    """

    spec = {
        "distributed": {
            "v1": DistributedV1Spec,
        },
        "unknown": {
            "v0": V0Spec,
        },
    }

    @staticmethod
    def load_spec(yaml_file, *args, **kwargs):
        debug = kwargs["debug"]
        spec = _ingest_yaml_file(yaml_file, debug)
        kind = spec.get("kind", "unknown")
        assert (
            kind in spec.keys()
        ), f"Kind {kind} is unknown. Kind should be one of {spec.keys()}"
        version = spec.get("version", "v0")
        assert (
            version in spec[kind].keys()
        ), f"Verion {version} is unknown. Kind should be one of {spec[kind].keys()}"
        return OperatorSpecFactory.spec[kind][version](spec, *args, **kwargs)


def _ingest_yaml_file(yaml_file, debug=False):
    if yaml_file:
        if os.path.exists(yaml_file):
            with open(yaml_file, "r") as f:
                spec = suppress_traceback(debug)(yaml.safe_load)(f.read())
        else:
            raise FileNotFoundError(f"{yaml_file} is not found")
    else:
        spec = {}
    return spec



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import os

import click

from ads.opctl.utils import publish_image as publish_image_cmd
from ads.opctl.utils import suppress_traceback
from ads.opctl.distributed.cmds import update_ini, docker_build_cmd
from ads.opctl.distributed.cmds import (
    initialize_workspace,
    show_config_info,
    cancel_distributed_run,
)
from configparser import ConfigParser


@click.group("distributed-training")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.option(
    "--framework",
    "-f",
    help="Distributed training framework type",
    type=click.Choice(
        ["dask", "horovod-tensorflow", "horovod-pytorch", "pytorch", "tensorflow"]
    ),
    default=None,
    required=True,
)
@click.option(
    "--version",
    help="version number of the framework resource",
    default="v1",
    required=False,
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.help_option("--help", "-h")
def init(framework, version, debug):
    """
    Initialize workspace to build distributed framework compatible container images.
    Supported frameworks are - dask, horovod-tensorflow, horovod-pytorch pytorch and
    tensorflow.
    """
    initialize_workspace(
        framework,
        version,
    )


@commands.command()
@click.option(
    "--job-id",
    "-j",
    help="Job OCID of the cluster",
    default=None,
    required=False,
)
@click.option(
    "--work-dir",
    "-f",
    help="Working directory provided when the cluster was created.",
    default=None,
    required=False,
)
@click.option(
    "--file",
    "-f",
    help="Yaml file with cluster run information",
    default=None,
    required=False,
)
@click.option(
    "--worker-info", help="Show worker information", is_flag=True, default=False
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.option(
    "--oci-config",
    help="OCI Config file location",
    default=None,
    required=False,
)
@click.option(
    "--oci-profile",
    help="oci config profile",
    default=None,
    required=False,
)
def show_config(job_id, work_dir, file, worker_info, debug, **kwargs):
    """
    Displays the config json files create by the main node and the worker nodes. By default only the config generated by the Main node is displayed.
    """
    if not (job_id and work_dir) and not file:
        print("Either job ocid or yaml file with `jobId` key is required")
        return
    show_config_info(
        job_id,
        work_dir,
        file,
        worker_info,
        **kwargs,
    )


@commands.command()
@click.option(
    "--job-id",
    "-j",
    help="Job OCID of the cluster",
    default=None,
    required=False,
)
@click.option(
    "--file",
    "-f",
    help="Yaml file with cluster run information",
    default=None,
    required=False,
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.option(
    "--oci-config",
    help="OCI Config file location",
    default=None,
    required=False,
)
@click.option(
    "--oci-profile",
    help="oci config profile",
    default=None,
    required=False,
)
def cancel(job_id, file, debug, **kwargs):
    """Terminate the running cluster. Use Job ID of the main jobrun or use the file which has the yaml output displayed when cluster was created using ``ads opctl run -f <yaml>``"""
    if not (job_id or file):
        print("Either job ocid or yaml file with `jobId` key is required")
        return
    cancel_distributed_run(job_id, file, **kwargs)


@commands.command()
@click.option("--tag", "-t", help="tag of image", required=False, default=None)
@click.option(
    "--registry", "-reg", help="registry to push to", required=False, default=None
)
@click.option(
    "--dockerfile",
    "-df",
    help="relative path to Dockerfile",
    required=False,
    default=None,
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=None)
@click.option(
    "--push", "-p", help="push docker image to registry", is_flag=True, default=None
)
@click.option(
    "--source-folder",
    "-s",
    help="source scripts folder that will be mounted during a local run",
    required=False,
    default=None,
)
def build_image(tag, registry, dockerfile, source_folder, **kwargs):
    debug = kwargs.pop("debug")
    push = kwargs["push"]
    ini = update_ini(
        tag, registry, dockerfile, source_folder, config=None, nobuild=False
    )
    img_name = ini.get("main", "registry") + ":" + ini.get("main", "tag")
    docker_build_cmd(ini)
    if push:
        print("pushing docker image")
        suppress_traceback(debug)(publish_image_cmd)(img_name)


@commands.command()
@click.argument("image", default="None")
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def publish_image(**kwargs):
    debug = kwargs.pop("debug")
    if kwargs["image"] == "None":
        config = ConfigParser()
        if os.path.isfile("config.ini"):
            config.read("config.ini")
            img_name = config.get("main", "registry") + ":" + config.get("main", "tag")
        else:
            raise ValueError("Arg image is missing")
    else:
        img_name = kwargs["image"]
    suppress_traceback(debug)(publish_image_cmd)(img_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import dataclass, field
from typing import Dict

from ads.common.serializer import DataClassSerializable
from ads.model.runtime.env_info import TrainingEnvInfo
from ads.model.runtime.utils import MODEL_PROVENANCE_SCHEMA_PATH, SchemaValidator


@dataclass(repr=False)
class TrainingCode(DataClassSerializable):
    """TrainingCode class."""

    artifact_directory: str = ""

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        assert obj_dict and (
            "ARTIFACT_DIRECTORY" in obj_dict
        ), "`training_code` must have `ARTIFACT_DIRECTORY` field."
        return True


@dataclass(repr=False)
class ModelProvenanceDetails(DataClassSerializable):
    """ModelProvenanceDetails class."""

    project_ocid: str = ""
    tenancy_ocid: str = ""
    training_code: TrainingCode = field(default_factory=TrainingCode)
    training_compartment_ocid: str = ""
    training_conda_env: TrainingEnvInfo = field(default_factory=TrainingEnvInfo)
    training_region: str = ""
    training_resource_ocid: str = ""
    user_ocid: str = ""
    vm_image_internal_id: str = ""

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """validate the yaml file.

        Parameters
        ----------
        obj_dict: (Dict)
            yaml file content to validate.

        Returns
        -------
        bool
            Validation result.
        """
        validator = SchemaValidator(schema_file_path=MODEL_PROVENANCE_SCHEMA_PATH)
        return validator.validate(obj_dict)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import click
from ads.common.auth import AuthType, OCIAuthContext
from ads.opctl.utils import suppress_traceback
from ads.pipeline.ads_pipeline import Pipeline
from ads.pipeline.ads_pipeline_run import LogType, PipelineRun


@click.group("pipeline")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.help_option("--help", "-h")
@click.option("--file", "-f", help="Path to YAML file", required=True)
@click.option(
    "--auth",
    "-a",
    help="Authentication method",
    type=click.Choice(AuthType.values()),
    default=AuthType.API_KEY,
)
@click.option(
    "--oci-profile", "-p", help="OCI config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="Set debug mode", is_flag=True, default=False)
def run(file, auth, oci_profile, debug):
    suppress_traceback(debug)(_run)(file, auth, oci_profile)


def _run(file, auth, oci_profile):
    if auth == AuthType.RESOURCE_PRINCIPAL:
        oci_profile = None
    with OCIAuthContext(profile=oci_profile):
        pipeline = Pipeline.from_yaml(uri=file)
        pipeline.create()
        print(pipeline.id)
        pipeline_run = pipeline.run()
        print(pipeline_run.id)


@commands.command()
@click.help_option("--help", "-h")
@click.argument("ocid")
@click.option(
    "--auth",
    "-a",
    help="Authentication method",
    type=click.Choice(AuthType.values()),
    default=AuthType.API_KEY,
)
@click.option(
    "--oci-profile", "-p", help="OCI config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="Set debug mode", is_flag=True, default=False)
@click.option(
    "--delete-related-job-runs",
    "-j",
    help="Whether to delete related Job runs or not.",
    is_flag=True,
    default=True,
)
def delete(ocid, auth, oci_profile, debug, delete_related_job_runs):
    suppress_traceback(debug)(_delete)(ocid, auth, oci_profile, delete_related_job_runs)


def _delete(ocid, auth, oci_profile, delete_related_job_runs):
    if auth == AuthType.RESOURCE_PRINCIPAL:
        oci_profile = None
    with OCIAuthContext(profile=oci_profile):
        run = PipelineRun.from_ocid(ocid)
        run.delete(delete_related_job_runs=delete_related_job_runs)


@commands.command()
@click.help_option("--help", "-h")
@click.argument("ocid")
@click.option(
    "--log-type",
    "-l",
    help="Log type",
    type=click.Choice(LogType.values()),
    default=LogType.CUSTOM_LOG,
)
@click.option(
    "--steps",
    "-s",
    help="Pipleine steps to monitor",
    default=None,
)
@click.option(
    "--auth",
    "-a",
    help="Authentication method",
    type=click.Choice(AuthType.values()),
    default=AuthType.API_KEY,
)
@click.option(
    "--oci-profile", "-p", help="OCI config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="Set debug mode", is_flag=True, default=False)
def watch(ocid, log_type, steps, auth, oci_profile, debug):
    suppress_traceback(debug)(_watch)(ocid, log_type, steps, auth, oci_profile)


def _watch(ocid, log_type, steps, auth, oci_profile):
    if auth == AuthType.RESOURCE_PRINCIPAL:
        oci_profile = None
    with OCIAuthContext(profile=oci_profile):
        run = PipelineRun.from_ocid(ocid)
        run.watch(steps=steps, log_type=log_type)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import dataclass, field
from typing import List


class WrongEntityFormatLabelNotString(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Label is not a String.")


class WrongEntityFormatOffsetNotInteger(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Offset is not an Integer.")


class WrongEntityFormatLengthNotInteger(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Length is not an Integer.")


class WrongEntityFormatOffsetIsNegative(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Offset is not a Nonnegative.")


class WrongEntityFormatLengthIsNegative(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Length is not a Nonnegative.")


class WrongEntityFormatLabelIsEmpty(ValueError):
    def __init__(self):
        super().__init__("Wrong entity format. Label is empty.")


@dataclass
class NERItem:
    """NERItem class which is a representation of a token span.

    Attributes
    ----------
    label: str
        Entity name.
    offset: int
        The token span's entity start index position in the text.
    length: int
        Length of the token span.
    """

    label: str = ""
    offset: int = 0
    length: int = 0

    def _validate(self):
        """Validates the instance.

        Raises
        ------
        WrongEntityFormat
           If the entity has a wrong format.

        """
        if not isinstance(self.label, str):
            raise WrongEntityFormatLabelNotString()
        if not (
            isinstance(self.offset, int)
            or (isinstance(self.offset, float) and self.offset.is_integer())
        ):
            raise WrongEntityFormatOffsetNotInteger()
        if not (
            isinstance(self.length, int)
            or (isinstance(self.length, float) and self.length.is_integer())
        ):
            raise WrongEntityFormatLengthNotInteger()
        if self.offset < 0:
            raise WrongEntityFormatOffsetIsNegative()
        if self.length < 0:
            raise WrongEntityFormatLengthIsNegative()
        if self.label == "":
            raise WrongEntityFormatLabelIsEmpty()

    def __post_init__(self):
        self._validate()

    def to_spacy(self) -> tuple:
        """Converts one NERItem to the spacy format.

        Returns
        -------
        Tuple
            NERItem in the spacy format
        """
        return (self.offset, self.offset + self.length, self.label)

    @classmethod
    def from_spacy(cls, token) -> "NERItem":
        offset, end, label = token
        return cls(label, offset, end - offset)


@dataclass
class NERItems:
    """NERItems class consists of a list of NERItem.

    Attributes
    ----------
    items: List[NERItem]
        List of NERItem.
    """

    items: List[NERItem] = field(default_factory=list)

    def __getitem__(self, index: int) -> NERItem:
        return self.items[index]

    def to_spacy(self) -> List[tuple]:
        """Converts NERItems to the spacy format.

        Returns
        -------
        List[tuple]
            List of NERItems in the Spacy format.
        """
        return [item.to_spacy() for item in self.items]



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import os
import re
from dataclasses import dataclass
from typing import Dict
from urllib.parse import urlparse

from ads.common import auth as authutil
from ads.common import oci_client


class InvalidObjectStoragePath(Exception):
    """Invalid Object Storage Path."""

    pass


@dataclass
class ObjectStorageDetails:
    """Class that represents the Object Storage bucket URI details.

    Attributes
    ----------
    bucket: str
        The Object Storage bucket name.
    namespace: (str, optional). Defaults to empty string.
        The Object Storage namespace. Will be extracted automatically if not provided.
    filepath: (str, optional). Defaults to empty string.
        The path to the object.
    auth: (Dict, optional). Defaults to None.
        ADS auth dictionary for OCI authentication.
        This can be generated by calling ads.common.auth.api_keys() or ads.common.auth.resource_principal()
        If this is None, ads.common.default_signer() will be used.
    """

    bucket: str
    namespace: str = ""
    filepath: str = ""
    auth: Dict = None

    def __post_init__(self):
        if not self.auth:
            self.auth = authutil.default_signer()
        # Extract OS namespace if not provided.
        if not self.namespace:
            os_client = oci_client.OCIClientFactory(**self.auth).object_storage
            self.namespace = os_client.get_namespace().data

    def __repr__(self):
        return self.path

    @property
    def path(self):
        """Full object storage path of this file."""
        return os.path.join(
            "oci://",
            self.bucket + "@" + self.namespace,
            self.filepath.lstrip("/") if self.filepath else "",
        )

    @classmethod
    def from_path(cls, env_path: str) -> "ObjectStorageDetails":
        """Construct an ObjectStorageDetails instance from conda pack path.

        Parameters
        ----------
        env_path: (str)
            codna pack object storage path.

        Raises
        ------
        Exception: OCI conda url path not properly configured.

        Returns
        -------
        ObjectStorageDetails
            An ObjectStorageDetails instance.
        """
        try:
            url_parse = urlparse(env_path)
            bucket_name = url_parse.username
            namespace = url_parse.hostname
            object_name = url_parse.path.strip("/")
            return cls(bucket=bucket_name, namespace=namespace, filepath=object_name)
        except:
            raise Exception(
                "OCI path is not properly configured. "
                "It should follow the pattern `oci://<bucket-name>@<namespace>/object_path`."
            )

    def to_tuple(self):
        """Returns the values of the fields of ObjectStorageDetails class."""
        return self.bucket, self.namespace, self.filepath

    def fetch_metadata_of_object(self) -> Dict:
        """Fetches the manifest metadata from the object storage of a conda pack.

        Returns
        -------
        Dict
            The metadata in dictionary format.
        """
        os_client = oci_client.OCIClientFactory(**self.auth).object_storage
        res = os_client.get_object(self.namespace, self.bucket, self.filepath)
        metadata = res.data.headers["opc-meta-manifest"]
        metadata_json = json.loads(metadata)
        return metadata_json

    @staticmethod
    def is_valid_uri(uri: str) -> bool:
        """Validates the Object Storage URI."""
        if not re.match(r"oci://*@*", uri):
            raise InvalidObjectStoragePath(
                f"The `{uri}` is not a valid Object Storage path. "
                "It must follow the pattern `oci://<bucket_name>@<namespace>/<prefix>`."
            )
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import oci.ai_language
import pandas as pd

from ads.common import auth as authutil
from ads.common import oci_client


class OCILanguage(object):  # pragma: no cover
    def __init__(self, auth=None):
        auth = auth if auth else authutil.default_signer()
        self.ai_client = oci_client.OCIClientFactory(**auth).ai_language

    @property
    def ner(self) -> pd.DataFrame:
        detect_language_entities_details = (
            oci.ai_language.models.DetectLanguageEntitiesDetails(text=self.string)
        )
        output = self.ai_client.detect_language_entities(
            detect_language_entities_details
        )
        entities = json.loads(str(output.data))["entities"]
        df = pd.DataFrame.from_dict(entities)
        df.columns = ["PII", "Length", "Offset", "Score", "Entity", "Type"]
        return df

    @property
    def language_dominant(self) -> pd.DataFrame:
        detect_dominant_language_details = (
            oci.ai_language.models.DetectDominantLanguageDetails(text=self.string)
        )
        output = self.ai_client.detect_dominant_language(
            detect_dominant_language_details
        )
        languages = json.loads(str(output.data))["languages"]
        df = pd.json_normalize(languages)
        df.columns = ["Code", "Language", "Score"]
        return df

    @property
    def key_phrase(self) -> pd.DataFrame:
        detect_language_key_phrases_details = (
            oci.ai_language.models.DetectLanguageKeyPhrasesDetails(text=self.string)
        )
        output = self.ai_client.detect_language_key_phrases(
            detect_language_key_phrases_details
        )
        key_phrase = json.loads(str(output.data))["key_phrases"]
        df = pd.DataFrame.from_dict(key_phrase)
        df.columns = ["Score", "Text"]
        return df

    @property
    def absa(self) -> pd.DataFrame:
        detect_language_sentiments_details = (
            oci.ai_language.models.DetectLanguageSentimentsDetails(text=self.string)
        )
        output = self.ai_client.detect_language_sentiments(
            detect_language_sentiments_details
        )
        aspects = json.loads(str(output.data))["aspects"]
        df = pd.json_normalize(aspects)
        df.columns = [
            "Length",
            "Offset",
            "Sentiment",
            "Text",
            "Negative",
            "Neutral",
            "Positive",
        ]
        return df

    @property
    def text_classification(self) -> pd.DataFrame:
        detect_language_text_clf_details = (
            oci.ai_language.models.DetectLanguageTextClassificationDetails(
                text=self.string
            )
        )
        output = self.ai_client.detect_language_text_classification(
            detect_language_text_clf_details
        )
        text_classification = json.loads(str(output.data))["text_classification"]
        df = pd.DataFrame.from_dict(text_classification)
        df.columns = ["Label", "Score"]
        return df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import warnings

warnings.warn(
    (
        "The `ads.common.model_artifact` is deprecated in `oracle-ads 2.6.9` and will be removed in `oracle-ads 3.0`."
        "Use framework specific Model utility class for saving and deploying model. "
        "Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html"
    ),
    DeprecationWarning,
    stacklevel=2,
)

import fnmatch
import importlib
import json
import os
import re
import git
import shutil
import subprocess
import sys
import textwrap
import uuid
import python_jsonschema_objects as pjs
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, Union

import ads.dataset.factory as factory
import fsspec
import numpy as np
import oci.data_science
import oci.exceptions
import pandas as pd
import pkg_resources
import yaml

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import logger, utils
from ads.common import auth as authutil
from ads.common.data import ADSData
from ads.common.error import ChangesNotCommitted
from ads.model.model_introspect import (
    TEST_STATUS,
    Introspectable,
    IntrospectionNotPassed,
    ModelIntrospect,
)
from ads.model.model_metadata import (
    METADATA_SIZE_LIMIT,
    MetadataCustomCategory,
    MetadataCustomKeys,
    MetadataSizeTooLarge,
    MetadataTaxonomyKeys,
    ModelCustomMetadata,
    ModelCustomMetadataItem,
    ModelTaxonomyMetadata,
    UseCaseType,
)
from ads.common.object_storage_details import (
    InvalidObjectStoragePath,
    ObjectStorageDetails,
)
from ads.common.utils import DATA_SCHEMA_MAX_COL_NUM
from ads.config import (
    JOB_RUN_COMPARTMENT_OCID,
    JOB_RUN_OCID,
    NB_SESSION_COMPARTMENT_OCID,
    NB_SESSION_OCID,
    PROJECT_OCID,
)
from ads.common.decorator.deprecate import deprecated
from ads.feature_engineering.schema import DataSizeTooWide, Schema, SchemaSizeTooLarge
from ads.model.extractor.model_info_extractor_factory import ModelInfoExtractorFactory
from ads.model.model_version_set import ModelVersionSet
from ads.model.common.utils import fetch_manifest_from_conda_location
from git import InvalidGitRepositoryError, Repo

from oci.data_science.models import ModelProvenance

try:
    from yaml import CDumper as dumper
    from yaml import CLoader as loader
except:
    from yaml import Dumper as dumper
    from yaml import Loader as loader

MODEL_ARTIFACT_VERSION = "3.0"
INPUT_SCHEMA_FILE_NAME = "input_schema.json"
OUTPUT_SCHEMA_FILE_NAME = "output_schema.json"

_TRAINING_RESOURCE_OCID = JOB_RUN_OCID or NB_SESSION_OCID
_COMPARTMENT_OCID = NB_SESSION_COMPARTMENT_OCID or JOB_RUN_COMPARTMENT_OCID


class InvalidDataType(Exception):
    """Invalid Data Type."""

    pass


SAMPLE_RUNTIME_YAML = f"""
MODEL_ARTIFACT_VERSION: '{MODEL_ARTIFACT_VERSION}'
MODEL_DEPLOYMENT:
    INFERENCE_CONDA_ENV:
        INFERENCE_ENV_SLUG: <slug of the conda environment>
        INFERENCE_ENV_TYPE: <data_science or published>
        INFERENCE_ENV_PATH: oci://<bucket-name>@<namespace>/<prefix>/<env>.tar.gz
        INFERENCE_PYTHON_VERSION: <python version>
"""


class ConflictStrategy(object):
    IGNORE = "IGNORE"
    UPDATE = "UPDATE"
    CREATE = "CREATE"


class PACK_TYPE(Enum):
    SERVICE_PACK = "data_science"
    USER_CUSTOM_PACK = "published"


class ModelArtifact(Introspectable):
    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def __init__(
        self,
        artifact_dir,
        conflict_strategy=ConflictStrategy.IGNORE,
        install_libs=False,
        reload=True,
        create=False,
        progress=None,
        model_file_name="model.onnx",
        inference_conda_env=None,
        data_science_env=False,
        ignore_deployment_error=False,
        inference_python_version=None,
    ):
        """
        A class used to construct model artifacts.

        ...

        Attributes
        ----------
        artifact_dir: str
            Path to the model artifacts.
        conflict_strategy: ConflictStrategy, default: IGNORE
            How to handle version conflicts between the current environment and the requirements of
            model artifact.
        install_libs: bool
            Re-install the environment inwhich the model artifact were trained in.
        reload: bool
            Reload the model into the environment.
        create: bool
            Create the `runtime.yaml` file.
        progress:
            Show a progress bar.
        model_file_name: str
            Name of the model file.
        inference_conda_env: str
            The inference conda environment. If provided, the value will be set in the runtime.yaml.
            This is expected to be full oci URI format - `oci://{bucket}@{namespace}/path/to/condapack`.
        data_science_env: bool
            Is the inference conda environment managed by the Oracle Data Science service?
        ignore_deployment_error: bool
            Determine whether to turn off logging for deployment error.
            If set to True, the `.prepare()` method will ignore errors that impact model deployment.
        inference_python_version: str Optional, default None
            The version of Python to be used in inference. The value will be set in the `runtime.yaml` file


        Methods
        -------
        reload(self, model_file_name=None)
            Reload the files in the model artifact directory.
        verify(self, input_data)
            Verifies a model artifact directory.
        install_requirements(self, conflict_strategy=ConflictStrategy.IGNORE)
            Installs missing libraries listed in the model artifact.
        populate_metadata(self, model=None, use_case_type=None)
            Extracts and populate taxonomy metadata from the model.
        save(
            self,
            display_name: str = None,
            description: str = None,
            project_id: str = None,
            compartment_id: str = None,
            training_script_path: str = None,
            ignore_pending_changes: bool = False,
            auth: dict = None,
            training_id: str = None,
            timeout: int = None,
            ignore_introspection=False,
        )
            Saves this model artifact in model catalog.
        populate_schema(
            self,
            data_sample: ADSData = None,
            X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
            y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        )
            Populates input and output schema.
        introspect(self) -> pd.DataFrame
            Runs model introspection.

        """
        self.artifact_dir = (
            artifact_dir[:-1] if artifact_dir.endswith("/") else artifact_dir
        )
        self._introspect = ModelIntrospect(self)
        self.model = None
        self.score = None
        self.inference_conda_env = inference_conda_env
        self.data_science_env = data_science_env
        self.ignore_deployment_error = ignore_deployment_error
        self.metadata_taxonomy = ModelTaxonomyMetadata()
        self.metadata_custom = ModelCustomMetadata()
        self.schema_input = Schema()
        self.schema_output = Schema()
        self._serialization_format = None
        self.inference_python_version = inference_python_version

        if create:
            self.progress = progress
            if "CONDA_PREFIX" in os.environ and "NB_SESSION_OCID" in os.environ:
                self._generate_runtime_yaml(model_file_name=model_file_name)
            else:
                self._generate_empty_runtime_yaml(
                    model_file_name=model_file_name,
                    data_science_env=data_science_env,
                    inference_conda_env=inference_conda_env,
                    inference_python_version=inference_python_version,
                )
            self.version = MODEL_ARTIFACT_VERSION

        # This will re-install the environment inwhich the model artifact was trained in.
        if install_libs:
            self.install_requirements(conflict_strategy=conflict_strategy)
        # This will reload the model into the environment
        if reload:
            self.reload(model_file_name=model_file_name)

    def __repr__(self):
        res = "Artifact directory: %s\n" % self.artifact_dir
        res += "Contains: %s" % str(self._get_files())
        return res

    def __getattr__(self, item):
        return getattr(self.score, item)

    def __fetch_repo_details(self, training_code_info):
        repo = git.Repo(".", search_parent_directories=True)
        # get repository url
        if len(repo.remotes) > 0:
            repository_url = (
                repo.remotes.origin.url
                if repo.remotes.origin in repo.remotes
                else list(repo.remotes.values())[0].url
            )
        else:
            repository_url = "file://" + repo.working_dir  # no remote repo

        git_branch = None
        git_commit = None
        try:
            # get git branch
            git_branch = format(repo.active_branch)
            # get git commit
            git_commit = format(str(repo.head.commit.hexsha))
            training_code_info.GIT_COMMIT = git_commit
        except ValueError:
            # do not set commit if there isn't any
            pass

        training_code_info.GIT_REMOTE = repository_url
        training_code_info.GIT_BRANCH = git_branch
        training_code_info.ARTIFACT_DIRECTORY = self.artifact_dir
        return repo, training_code_info

    def __fetch_training_env_details(self, training_info):
        conda_prefix = os.environ.get("CONDA_PREFIX", None)
        pack_name = "NOT FOUND"
        try:
            manifest = fetch_manifest_from_conda_location(conda_prefix)
            manifest_type = manifest["type"]
            pack_name = manifest["pack_path"] if "pack_path" in manifest else None
            slug = manifest["slug"] if "slug" in manifest else ""

            if manifest_type == PACK_TYPE.USER_CUSTOM_PACK.value:
                if os.path.exists(
                    os.path.join(os.path.expanduser("~"), "conda", "config.yaml")
                ):
                    with open(
                        (os.path.join(os.path.expanduser("~"), "conda", "config.yaml"))
                    ) as conf:
                        user_config = yaml.load(conf, Loader=yaml.FullLoader)
                    pack_bucket = user_config["bucket_info"]["name"]
                    pack_namespace = user_config["bucket_info"]["namespace"]
                else:
                    logger.warning(
                        f"Cannot resolve the bucket name or namespace for the conda environment {conda_prefix}. "
                        f"You can set these values while saving the model or run `odsc init -b *bucket-name* -n *namespace*` and rerun the prepare step again."
                    )
            if not manifest_type or manifest_type.lower() not in [
                PACK_TYPE.USER_CUSTOM_PACK.value,
                PACK_TYPE.SERVICE_PACK.value,
            ]:
                if not self.ignore_deployment_error:
                    raise Exception(
                        f"Unknown manifest type. Manifest Type: {manifest_type or 'None'}"
                    )
            if not pack_name:
                if manifest_type == PACK_TYPE.USER_CUSTOM_PACK.value:
                    if self.data_science_env:
                        raise Exception(
                            f"For Published conda environments, assign the path of the environment in "
                            + "Object Storage to the `inference_conda_env` parameter and set the "
                            + "parameter `data_science_env` to `False`."
                        )
                    error_message = (
                        f"Pack destination is not known from the manifest file in {conda_prefix}. "
                        + "If it was cloned from another environment, consider publishing it before "
                        + "preparing the model artifact."
                    )
                    if self.ignore_deployment_error:
                        logger.warn(error_message)
                    else:
                        if not self.inference_conda_env:
                            logger.error(error_message)
                            logger.info(
                                "Provide a URI to the conda environment that you wish to use with the model "
                                "deployment service if you do not want to publish the current training environment."
                            )
                            raise Exception(
                                f"Could not resolve the path in the Object Storage for the conda environment: {conda_prefix}"
                            )
                else:
                    logger.warn(
                        f"Could not resolve the Object Storage destination of {conda_prefix}. Correct "
                        "the environment name and Object Storage details when saving."
                    )
        except Exception as e:
            raise e
        training_info.TRAINING_ENV_SLUG = slug
        if manifest_type.lower() in [
            PACK_TYPE.USER_CUSTOM_PACK.value,
            PACK_TYPE.SERVICE_PACK.value,
        ]:
            training_info.TRAINING_ENV_TYPE = manifest_type
        if pack_name:
            training_info.TRAINING_ENV_PATH = pack_name
        training_info.TRAINING_PYTHON_VERSION = sys.version.split("|")[0].strip()
        return training_info

    def __environment_details(self, model_provenance):
        model_provenance.TRAINING_REGION = os.environ.get("NB_REGION", "NOT_FOUND")
        model_provenance.TRAINING_COMPARTMENT_OCID = os.environ.get(
            "NB_SESSION_COMPARTMENT_OCID", "NOT_FOUND"
        )
        model_provenance.TRAINING_RESOURCE_OCID = os.environ.get(
            "NB_SESSION_OCID", "NOT_FOUND"
        )
        model_provenance.PROJECT_OCID = os.environ.get("PROJECT_OCID", "NOT_FOUND")
        model_provenance.TENANCY_OCID = os.environ.get("TENANCY_OCID", "NOT_FOUND")
        model_provenance.USER_OCID = os.environ.get("USER_OCID", "NOT_FOUND")
        model_provenance.VM_IMAGE_INTERNAL_ID = os.environ.get("VM_ID", "VMIDNOTSET")
        return model_provenance

    def __fetch_runtime_schema__(self):
        schema = None
        with open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "model_artifact_schema.json"
            )
        ) as schema_file:
            schema = json.load(schema_file)

        if not schema:
            raise Exception(
                "Cannot load schema file to generate the runtime.yaml file."
            )

        builder = pjs.ObjectBuilder(schema)
        ns = builder.build_classes()
        return ns

    def _generate_empty_runtime_yaml(
        self,
        model_file_name="model.onnx",
        data_science_env=False,
        inference_conda_env=None,
        inference_python_version=None,
    ):
        if self.progress:
            self.progress.update("Creating runtime.yaml configuration.")
        logger.warning(
            "Generating runtime.yaml template. This file needs to be updated "
            "before saving it to the model catalog."
        )
        content = yaml.load(SAMPLE_RUNTIME_YAML, Loader=yaml.FullLoader)
        print(
            f"The inference conda environment is {inference_conda_env} and the Python version is {inference_python_version}."
        )
        if inference_conda_env:
            content["MODEL_DEPLOYMENT"]["INFERENCE_CONDA_ENV"][
                "INFERENCE_ENV_SLUG"
            ] = ""
            content["MODEL_DEPLOYMENT"]["INFERENCE_CONDA_ENV"][
                "INFERENCE_ENV_TYPE"
            ] = ""
            content["MODEL_DEPLOYMENT"]["INFERENCE_CONDA_ENV"][
                "INFERENCE_ENV_PATH"
            ] = inference_conda_env
        if inference_python_version:
            content["MODEL_DEPLOYMENT"]["INFERENCE_CONDA_ENV"][
                "INFERENCE_PYTHON_VERSION"
            ] = str(inference_python_version)

        content["MODEL_DEPLOYMENT"]["INFERENCE_CONDA_ENV"]["INFERENCE_ENV_TYPE"] = (
            PACK_TYPE.SERVICE_PACK.value
            if data_science_env
            else PACK_TYPE.USER_CUSTOM_PACK.value
        )

        with open(os.path.join(self.artifact_dir, "runtime.yaml"), "w") as outfile:
            yaml.dump(content, outfile)

    def _generate_runtime_yaml(self, model_file_name="model.onnx"):
        if self.progress:
            self.progress.update("Creating runtime.yaml configuration.")

        ns = self.__fetch_runtime_schema__()
        training_env_info = self.__fetch_training_env_details(ns.TrainingCondaEnv())
        model_provenance = self.__environment_details(ns.ModelProvenance())
        model_provenance.TRAINING_CONDA_ENV = training_env_info
        try:
            _, training_code_info = self.__fetch_repo_details(ns.TrainingCodeInfo())
            model_provenance.TRAINING_CODE = training_code_info
        except git.InvalidGitRepositoryError:
            pass

        if not training_env_info.TRAINING_ENV_PATH:
            logger.warning(
                "You did not publish the conda environment in which the madel was trained. Publishing the "
                "conda environment ensures that the exact training environment can be re-used later."
            )
        inference_info = ns.InferenceCondaEnv()
        if not self.inference_conda_env:
            message = "By default, the inference conda environment is the same as the training conda environment. Use the `inference_conda_env` parameter to override."
            if (
                training_env_info.TRAINING_ENV_TYPE
                and training_env_info.TRAINING_ENV_PATH
            ):
                logger.info(message)
                inference_info.INFERENCE_ENV_SLUG = training_env_info.TRAINING_ENV_SLUG
                inference_info.INFERENCE_ENV_TYPE = training_env_info.TRAINING_ENV_TYPE
                inference_info.INFERENCE_ENV_PATH = training_env_info.TRAINING_ENV_PATH
                inference_info.INFERENCE_PYTHON_VERSION = (
                    training_env_info.TRAINING_PYTHON_VERSION
                )
                self.conda_env = str(training_env_info.TRAINING_ENV_SLUG)
        else:
            self.conda_env = os.path.basename(str(self.inference_conda_env))
            if self.inference_conda_env.startswith("oci://"):
                inference_info.INFERENCE_ENV_PATH = self.inference_conda_env
                try:
                    metadata_json = ObjectStorageDetails.from_path(
                        env_path=self.inference_conda_env
                    ).fetch_metadata_of_object()
                    inference_info.INFERENCE_PYTHON_VERSION = metadata_json["python"]
                except:
                    if not self.inference_python_version:
                        if not training_env_info.TRAINING_PYTHON_VERSION:
                            raise Exception(
                                "The Python version was not specified."
                                "Pass in the Python version when preparing a model."
                            )
                        else:
                            logger.warning(
                                "The Python version could not be inferred from the conda environment. Defaulting to the Python "
                                "version that was used in training."
                            )
                            inference_info.INFERENCE_PYTHON_VERSION = (
                                training_env_info.TRAINING_PYTHON_VERSION
                            )
                    else:
                        inference_info.INFERENCE_PYTHON_VERSION = (
                            self.inference_python_version
                        )
            else:
                pass
        model_deployment_info = None
        if inference_info.INFERENCE_ENV_PATH:
            model_deployment_info = ns.ModelDeployment()
            model_deployment_info.INFERENCE_CONDA_ENV = inference_info

        if (
            not self.inference_conda_env
            and not self.data_science_env
            and inference_info.INFERENCE_ENV_TYPE == PACK_TYPE.SERVICE_PACK.value
            and training_env_info.TRAINING_ENV_PATH == inference_info.INFERENCE_ENV_PATH
        ):
            error_message = (
                f"The inference conda environment {training_env_info.TRAINING_ENV_SLUG} may have changed. "
                + "Publish the current conda environment or set the parameter `data_science_env` to `True` "
                + "in the `.prepare()` method."
            )
            if not self.ignore_deployment_error:
                raise Exception(error_message)
            else:
                logger.warning(error_message)

        if not inference_info.INFERENCE_ENV_PATH and not self.inference_conda_env:
            error_message = (
                f"The inference conda environment is missing. Set the `inference_conda_env` parameter "
                + "or publish the conda environment and run the `.prepare()` method."
            )
            if not self.ignore_deployment_error:
                raise Exception(error_message)
            else:
                logger.warn(error_message)

        if model_deployment_info:
            self._runtime_info = ns.ModelArtifactSchema(
                MODEL_ARTIFACT_VERSION=MODEL_ARTIFACT_VERSION,
                MODEL_PROVENANCE=model_provenance,
                MODEL_DEPLOYMENT=model_deployment_info,
            )
        else:
            self._runtime_info = ns.ModelArtifactSchema(
                MODEL_ARTIFACT_VERSION=MODEL_ARTIFACT_VERSION,
                MODEL_PROVENANCE=model_provenance,
            )

        with open(os.path.join(self.artifact_dir, "runtime.yaml"), "w") as outfile:
            outfile.write("# Model runtime environment\n")
            yaml.dump(self._runtime_info.as_dict(), outfile, default_flow_style=False)

    def reload(self, model_file_name: str = None):
        """
        Reloads files in model artifact directory.

        Parameters
        ----------
        model_file_name: str
            The model file name.
        """
        spec = importlib.util.spec_from_file_location(
            "score%s" % uuid.uuid4(), os.path.join(self.artifact_dir, "score.py")
        )
        score = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(score)
        self.score = score
        if os.path.exists(os.path.join(self.artifact_dir, "runtime.yaml")):
            if model_file_name:
                self.model = self.score.load_model(model_file_name)
            else:
                self.model = self.score.load_model()
            with open(os.path.join(self.artifact_dir, "runtime.yaml")) as runtime_file:
                runtime = yaml.load(runtime_file, Loader=yaml.FullLoader)
            self.version = runtime["MODEL_ARTIFACT_VERSION"]
            try:
                self.VM_ID = runtime["MODEL_PROVENANCE"]["VM_IMAGE_INTERNAL_ID"]
            except KeyError:
                self.VM_ID = None
            try:
                self.conda_env = runtime["MODEL_PROVENANCE"]["TRAINING_CONDA_ENV"][
                    "TRAINING_ENV_SLUG"
                ]
            except KeyError:
                self.conda_env = None
        elif os.path.exists(os.path.join(self.artifact_dir, "ds-runtime.yaml")):
            self.model = self.score.load_model()
            with open(
                os.path.join(self.artifact_dir, "ds-runtime.yaml")
            ) as runtime_file:
                runtime = yaml.load(runtime_file, Loader=yaml.FullLoader)
            self.version = "1.0"
            self.VM_ID = None  # get ads/mlx version?
            self.conda_env = runtime["conda-env"]
        else:
            self.model = self.score.load_model()
            self.version = "0.0"
            self.VM_ID = "UNKNOWN"
            self.conda_env = "base"
            # raise FileNotFoundError(os.path.join(self.artifact_dir, 'runtime.yaml'))
        # __pycache__ was created during model_artifact.reload() above
        if os.path.exists(os.path.join(self.artifact_dir, "__pycache__")):
            shutil.rmtree(
                os.path.join(self.artifact_dir, "__pycache__"), ignore_errors=True
            )
        # extract model serialization format as part of custom metadata
        if model_file_name:
            self._serialization_format = self._extract_model_serialization_format(
                model_file_name
            )
            if (
                MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT
                in self.metadata_custom.keys
            ):
                self.metadata_custom[
                    MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT
                ].value = self._serialization_format
            else:
                self.metadata_custom.add(
                    key=MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT,
                    value=self._serialization_format,
                    description="The model serialization format",
                    category=MetadataCustomCategory.TRAINING_PROFILE,
                )

    @staticmethod
    def _extract_model_serialization_format(model_file_name):
        return os.path.splitext(model_file_name)[1][1:]

    def verify(self, input_data):
        """
        Verifies the contents of the  model artifact directory.

        Parameters
        ----------
        input_data : str, dict, BytesIO stream
            Data to be passed into the deployed model. It can be of type json (str), a dict object, or a BytesIO stream.
            All types get converted into a UTF-8 encoded BytesIO stream and is then sent to the handler.
            Any data handling past there is done in func.py. By default it looks for data
            under the keyword "input", and returns data under teh keyword "prediction".

        Returns
        -------
        output_data : the resulting prediction, formatted in the same way as input_data

        Example
         --------
         input_dict = {"input": train.X[:3].to_dict()}
         model_artifact.verify(input_dict)

         * returns {"prediction": [30/4, 24.8, 30.7]} *
        """

        # Fake Context obj created for Fn Handler
        class FakeCtx:
            def SetResponseHeaders(self, headers, status_code):
                return

        ctx = FakeCtx()
        from io import BytesIO

        if type(input_data) == str:
            data = BytesIO(input_data.encode("UTF-8"))
            data_type = "json"
        elif type(input_data) == dict:
            from json import dumps

            data = BytesIO(dumps(input_data).encode("UTF-8"))
            data_type = "dict"
        elif isinstance(type(input_data), type(BytesIO)):
            data = input_data
            data_type = "BytesIO"
        else:
            raise TypeError

        sys_path = sys.path.copy()
        try:
            if self.version.split(".")[0] not in ["0", "1"]:
                sys.path.insert(0, self.artifact_dir)
            else:
                sys.path.insert(0, os.path.join(self.artifact_dir, "fn-model"))
            import func

            resp = func.handler(ctx, data)
            output_json = resp.body()
        finally:
            # Reset in case func.py messes with it
            sys.path = sys_path

        if data_type == "json":
            return output_json
        output_bstream = BytesIO(resp.body().encode("UTF-8"))
        if data_type == "BytesIO":
            return output_bstream
        else:
            from json import load

            return load(output_bstream)

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def save(
        self,
        display_name: str = None,
        description: str = None,
        project_id: str = None,
        compartment_id: str = None,
        training_script_path: str = None,
        ignore_pending_changes: bool = False,
        auth: dict = None,
        training_id: str = None,
        timeout: int = None,
        ignore_introspection=True,
        freeform_tags=None,
        defined_tags=None,
        bucket_uri: Optional[str] = None,
        remove_existing_artifact: Optional[bool] = True,
        model_version_set: Optional[Union[str, ModelVersionSet]] = None,
        version_label: Optional[str] = None,
    ):
        """
        Saves the model artifact in the model catalog.

        Parameters
        ----------
        display_name : str, optional
            Model display name.
        description : str, optional
            Description for the model.
        project_id : str, optional
            Model's project OCID.
            If None, the default project OCID `config.PROJECT_OCID` would be used.
        compartment_id : str, optional
            Model's compartment OCID.
            If None, the default compartment OCID `config.NB_SESSION_COMPARTMENT_OCID` would be used.
        training_script_path : str, optional
            The training script path is either relative to the working directory,
            or an absolute path.
        ignore_pending_changes : bool, default: False
            If True, ignore uncommitted changes and use the current git HEAD commit for provenance metadata.
            This argument is used only when the function is called from a script in git managed directory.
        auth: dict
            Default is None. Default authetication is set using the `ads.set_auth()` method.
            Use the `ads.common.auth.api_keys()` or `ads.common.auth.resource_principal()` to create appropriate
            authentication signer and kwargs required to instantiate a DataScienceClient object.
        training_id: str, optional
            The training OCID for the model.
        timeout: int, default: 10
            The connection timeout in seconds.
        ignore_introspection: bool, optional
            Ignore the result of model introspection .
            If set to True, the `.save()` will ignore all model introspection errors.
        freeform_tags : dict(str, str), optional
            Freeform tags for the model.
        defined_tags : dict(str, dict(str, object)), optional
            Defined tags for the model.
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.
        model_version_set: (Union[str, ModelVersionSet], optional). Defaults to None.
            The Model version set OCID, or name, or `ModelVersionSet` instance.
        version_label: (str, optional). Defaults to None.
            The model version label.

        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> from ads.config import NB_SESSION_OCID

        >>> # Getting auth details.
        >>> # If you are using API keys
        >>> auth=ads.common.auth.api_keys()

        >>> # If you are using resource principal
        >>> auth=ads.common.auth.resource_principal()

        >>> # If you have set the auth type using ads.set_auth()
        >>> auth=ads.common.auth.default_signer()

        >>> # Preparing model artifacts
        >>> model_artifact = prepare_generic_model(
        ...     "path_to_model_artifacts",
        ...     force_overwrite=True,
        ...     data_science_env=True,
        ...     model=gamma_reg_model,
        ... )

        >>> # Saving model to the model catalog
        >>> model_artifact.save(
        ...     project_id=PROJECT_ID,
        ...     compartment_id=COMPARTMENT,
        ...     display_name="RF Classifier 2",
        ...     description="A sample Random Forest classifier",
        ...     ignore_pending_changes=True,
        ...     auth=auth,
        ...     training_id=NB_SESSION_OCID,
        ...     timeout=6000,
        ...     ignore_introspection = True
        ... )
        """
        if timeout and not isinstance(timeout, int):
            raise TypeError("Timeout must be an integer.")

        runtime_yaml_file = os.path.join(self.artifact_dir, "runtime.yaml")
        if os.path.exists(runtime_yaml_file):
            with open(runtime_yaml_file, "r") as mfile:
                runtime_prep_info = yaml.load(mfile, Loader=yaml.FullLoader)
                # runtime_info['pack-info'] = deployment_pack_info
        else:
            runtime_prep_info = {}
        ns = self.__fetch_runtime_schema__()
        runtime_info = ns.ModelArtifactSchema().from_json(json.dumps(runtime_prep_info))

        training_code_info = self._training_code_info(
            ns, training_script_path, ignore_pending_changes
        )
        if not training_id:
            training_id = _TRAINING_RESOURCE_OCID
        model_provenance_metadata = ModelProvenance(
            repository_url=str(training_code_info.GIT_REMOTE),
            git_branch=str(training_code_info.GIT_BRANCH),
            git_commit=str(training_code_info.GIT_COMMIT),
            script_dir=str(training_code_info.ARTIFACT_DIRECTORY),
            training_script=str(training_code_info.TRAINING_SCRIPT),
            training_id=training_id,
        )
        if getattr(runtime_info, "MODEL_PROVENANCE", None):
            runtime_info.MODEL_PROVENANCE.TRAINING_CODE = training_code_info

        logger.info(model_provenance_metadata)

        # handle the case where project_id and/or compartment_id is not specified by the user
        if not project_id and not PROJECT_OCID:
            raise ValueError("The `project_id` must be provided.")

        if not compartment_id and not NB_SESSION_COMPARTMENT_OCID:
            raise ValueError("The `compartment_id` must be provided.")

        if os.path.exists(os.path.join(self.artifact_dir, "__pycache__")):
            shutil.rmtree(
                os.path.join(self.artifact_dir, "__pycache__"), ignore_errors=True
            )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_ARTIFACTS,
                value=textwrap.shorten(
                    ", ".join(self._get_files()), 255, placeholder="..."
                ),
                description="The list of files located in artifacts folder.",
                category=MetadataCustomCategory.TRAINING_ENV,
            ),
            replace=True,
        )

        client_auth = auth if auth else authutil.default_signer()

        if timeout:
            if not client_auth.get("client_kwargs"):
                client_auth["client_kwargs"] = {}
            client_auth["client_kwargs"]["timeout"] = timeout

        if freeform_tags and not isinstance(freeform_tags, dict):
            raise TypeError("Freeform tags must be a dictionary.")

        if defined_tags and not isinstance(defined_tags, dict):
            raise TypeError("Defined tags must be a dictionary.")

        self._validate_metadata()
        self._validate_schema()

        with open(runtime_yaml_file, "w") as mfile:
            yaml.dump(runtime_info.as_dict(), mfile, Dumper=dumper)

        if not ignore_introspection:
            self._introspect()
            if self._introspect.status == TEST_STATUS.NOT_PASSED:
                msg = (
                    "Model introspection not passed. "
                    "Use `.introspect()` method to get detailed information and follow the "
                    "messages to fix it. To save model artifacts ignoring introspection "
                    "use `.save(ignore_introspection=True...)`."
                )
                raise IntrospectionNotPassed(msg)
        try:
            from ads.catalog.model import ModelCatalog

            return ModelCatalog(
                compartment_id=compartment_id,
                ds_client_auth=client_auth,
                identity_client_auth=client_auth,
            ).upload_model(
                self,
                provenance_metadata=model_provenance_metadata,
                display_name=display_name,
                description=description,
                project_id=project_id,
                freeform_tags=freeform_tags,
                defined_tags=defined_tags,
                bucket_uri=bucket_uri,
                remove_existing_artifact=remove_existing_artifact,
                model_version_set=model_version_set,
                version_label=version_label,
            )
        except oci.exceptions.RequestException as e:
            if "The write operation timed out" in str(e):
                logger.error(
                    "The save operation timed out. Try to set a longer timeout e.g. save(timeout=600, ...)."
                )

    def _validate_schema(self):
        if not self._validate_schema_size(self.schema_input, INPUT_SCHEMA_FILE_NAME):
            self.schema_input.to_json_file(
                os.path.join(self.artifact_dir, INPUT_SCHEMA_FILE_NAME)
            )
        if not self._validate_schema_size(self.schema_output, OUTPUT_SCHEMA_FILE_NAME):
            self.schema_output.to_json_file(
                os.path.join(self.artifact_dir, OUTPUT_SCHEMA_FILE_NAME)
            )
        self.schema_input.validate_schema()
        self.schema_output.validate_schema()

    def _validate_metadata(self):
        self.metadata_custom.validate()
        self.metadata_taxonomy.validate()
        total_size = self.metadata_custom.size() + self.metadata_taxonomy.size()
        if total_size > METADATA_SIZE_LIMIT:
            raise MetadataSizeTooLarge(total_size)
        return True

    def _training_code_info(
        self, ns, training_script_path=None, ignore_pending_changes=False
    ):
        try:
            repo, training_code_info = self.__fetch_repo_details(ns.TrainingCodeInfo())
        except git.InvalidGitRepositoryError:
            repo = None
            training_code_info = ns.TrainingCodeInfo()
        if training_script_path is not None:
            if not os.path.exists(training_script_path):
                logger.warning(
                    f"Training script {os.path.abspath(training_script_path)} does not exists."
                )
            else:
                training_script = os.path.abspath(training_script_path)
                self._assert_path_not_dirty(
                    training_script_path, repo, ignore_pending_changes
                )
                training_code_info.TRAINING_SCRIPT = training_script

        self._assert_path_not_dirty(self.artifact_dir, repo, ignore_pending_changes)
        training_code_info.ARTIFACT_DIRECTORY = os.path.abspath(self.artifact_dir)

        return training_code_info

    def _assert_path_not_dirty(self, path, repo, ignore):
        if repo is not None and not ignore:
            path_abs = os.path.abspath(path)
            if os.path.commonpath([path_abs, repo.working_dir]) == repo.working_dir:
                path_relpath = os.path.relpath(path_abs, repo.working_dir)
                if repo.is_dirty(path=path_relpath) or any(
                    [
                        os.path.commonpath([path_relpath, untracked]) == path_relpath
                        for untracked in repo.untracked_files
                    ]
                ):
                    raise ChangesNotCommitted(path_abs)

    def install_requirements(self, conflict_strategy=ConflictStrategy.IGNORE):
        """
        Installs missing libraries listed in the model artifacts.

        Parameters
        ----------
        conflict_strategy : ConflictStrategy, default: IGNORE
            Update the conflicting dependency to the version required by the model artifact.
            Valid values: "IGNORE" or ConflictStrategy.IGNORE, "UPDATE" or ConflictStrategy.UPDATE.
            IGNORE: Use the installed version in  case of a conflict.
            UPDATE: Force update dependency to the version required by model artifact in case of conflict.
        """
        importlib.reload(pkg_resources)
        from pkg_resources import DistributionNotFound, VersionConflict

        if self.version.split(".")[0] not in ["0", "1"] and os.path.exists(
            Path(os.path.join(self.artifact_dir), "requirements.txt")
        ):
            requirements = (
                Path(os.path.join(self.artifact_dir), "requirements.txt")
                .read_text()
                .strip()
                .split("\n")
            )
        elif self.version.split(".")[0] in ["0", "1"] and Path(
            os.path.join(self.artifact_dir), "ds-requirements.txt"
        ):
            requirements = (
                Path(os.path.join(self.artifact_dir), "ds-requirements.txt")
                .read_text()
                .strip()
                .split("\n")
            )
        else:
            raise FileNotFoundError(
                "Could not find requirements.txt. Install the necessary libraries and "
                "re-construct the model artifact with install_libs=False."
            )

        version_conflicts = {}
        for requirement in requirements:
            try:
                pkg_resources.require(requirement)
            except VersionConflict as vc:
                if conflict_strategy == ConflictStrategy.UPDATE:
                    pip_install("%s%s" % (vc.req.name, vc.req.specifier), "-U")
                elif conflict_strategy == ConflictStrategy.IGNORE:
                    version_conflicts[
                        "%s==%s" % (vc.dist.key, vc.dist.parsed_version)
                    ] = "%s%s" % (vc.req.name, vc.req.specifier)
            except DistributionNotFound as dnf:
                pip_install(requirement)
                # distributions_not_found.add('%s%s' % (dnf.req.name, dnf.req.specifier))
        if len(version_conflicts) > 0:
            print(
                "\033[93m"
                + str(VersionConflictWarning(version_conflicts=version_conflicts))
                + "\033[0m"
            )

    def _get_files(self):
        if os.path.exists(os.path.join(self.artifact_dir, ".model-ignore")):
            ignore_patterns = (
                Path(os.path.join(self.artifact_dir), ".model-ignore")
                .read_text()
                .strip()
                .split("\n")
            )
        else:
            ignore_patterns = []
        file_names = []
        for root, dirs, files in os.walk(self.artifact_dir):
            for name in files:
                file_names.append(os.path.join(root, name))
            for name in dirs:
                file_names.append(os.path.join(root, name))

        for ignore in ignore_patterns:
            if not ignore.startswith("#") and ignore.strip() != "":
                matches = []
                for file_name in file_names:
                    if ignore.endswith("/"):
                        ignore = ignore[:-1] + "*"
                    if not re.search(
                        fnmatch.translate("/%s" % ignore.strip()), file_name
                    ):
                        matches.append(file_name)
                file_names = matches
        return [
            matched_file[len(self.artifact_dir) + 1 :] for matched_file in file_names
        ]

    def _save_data_from_memory(
        self,
        prefix: str,
        train_data: Union[pd.DataFrame, list, np.ndarray],
        validation_data: Union[pd.DataFrame, list, np.ndarray] = None,
        train_data_name: str = "train.csv",
        validation_data_name: str = "validation.csv",
        storage_options: dict = None,
        **kwargs,
    ):
        """
        Save data to Object Storage.
        return [
            matched_file[len(self.artifact_dir) + 1 :] for matched_file in file_names
        ]

        Parameters
        ----------
        prefix: str
            A prefix to append to the Object Storage key.
            e.g. oci://bucket_name@namespace/prefix
        train_data: Union[pd.DataFrame, list, np.ndarray].
            The training data to be stored.
        validation_data: Union[pd.DataFrame, list, np.ndarray]. Default None
            The validation data to be stored.
        train_data_name: str. Default 'train.csv'.
            Filename used to save the train data. The key is prefix/train_data_name.
        validation_data_name: str. Default 'train.csv'.
            Filename used to save the validation data. The key is prefix/validation_data_name.
        storage_options: dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `storage_options` set using `DatasetFactory.set_default_storage()`.

        Returns
        -------
        None
            Nothing.
        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> import ocifs
        >>> import oci
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': 'ocid5.user.oc1..aaaaaaaab3geixlk***********************',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': 'ocid5.tenancy.oc1..aaaaaaaag*************************',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> path_to_generic_model_artifact = tempfile.mkdtemp()
        >>> df = pd.DataFrame([[1, 2], [2, 3], [3, 4], [4, 3]])
        >>> generic_model_artifact = prepare_generic_model(path_to_generic_model_artifact,
        ...                                   force_overwrite=True, data_science_env=True,
        ...                                   ignore_deployment_error=True)
        >>> generic_model_artifact._save_data_from_memory(prefix = 'oci://bucket_name@namespace/folder_name',
        ... train_data=df, storage_options=storage_options)
        """
        if not re.match(r"oci://*@*", prefix):
            raise InvalidObjectStoragePath(
                "`prefix` is not valid. It must have the pattern 'oci://bucket_name@namespace/key'."
            )
        if not storage_options:
            storage_options = factory.default_storage_options
        if not storage_options:
            storage_options = {"config": {}}

        self._save_from_memory(
            train_data, prefix, train_data_name, storage_options, "training", **kwargs
        )

        if validation_data is not None:
            self._save_from_memory(
                validation_data,
                prefix,
                validation_data_name,
                storage_options,
                "validation",
                **kwargs,
            )

    def _save_data_from_file(
        self,
        prefix: str,
        train_data_path: str = None,
        validation_data_path: str = None,
        storage_options: dict = None,
        **kwargs,
    ):
        """
        Save the data to Object Storage.

        Parameters
        ----------
        prefix: str
            The Object Storage prefix to store the data. When `train_data_path` or
            `validation_data_path` are provided, they are stored under this prefix
            with their original filenames. If the data are already stored on Object
            Storage, you can provide the path to the data. If no local data path is provided,
            no data is `prefix` is saved in the custom metadata.
        train_data_path: str. Default None.
            Local path for the training data.
        validation_data_path: str. Default None.
            Local path for the validation data.
        storage_options: dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `storage_options` set using `DatasetFactory.set_default_storage()`.

        Keyword Arguments
        _________________
        data_type:
            Either `training` or `validation`. Used when the data are
            already stored remotely and you want to record the path in
            `metadata_custom`. Pass the prefix of your data and `data_type`
            to indicate whether this data is of `training` or `validation` type.
            The `storage_options` is needed in this case.

        Returns
        -------
        None
            Nothing.

        Examples
        ________
        >>> from ads.common.model_artifact import ModelArtifact
        >>> import ocifs
        >>> import oci
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': 'ocid5.user.oc1..aaaaaaaab3geixlk***********************',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': 'ocid5.tenancy.oc1..aaaaaaaag*************************',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> path_to_generic_model_artifact = tempfile.mkdtemp()
        >>> generic_model_artifact = prepare_generic_model(path_to_generic_model_artifact,
        ...                                   force_overwrite=True, data_science_env=True,
        ...                                   ignore_deployment_error=True)
        >>> generic_model_artifact._save_data_from_file(oci_storage_path = 'oci://bucket_name@namespace/folder_name',
        ... train_data_path = '~/orcl_attrition*.csv', storage_options=storage_options)
        """
        if not re.match(r"oci://*@*", prefix):
            raise InvalidObjectStoragePath(
                "`prefix` is not valid. It must have the pattern 'oci://bucket_name@namespace/key'."
            )
        if not storage_options:
            storage_options = factory.default_storage_options
        if not storage_options:
            storage_options = {"config": {}}

        if train_data_path is not None:
            assert isinstance(train_data_path, str), "A path to the data is required."
            self._save_from_local_file(
                prefix=prefix,
                file_path=train_data_path,
                storage_options=storage_options,
                data_type="training",
            )
        if validation_data_path is not None:
            assert isinstance(
                validation_data_path, str
            ), "A path to the data is required."
            self._save_from_local_file(
                prefix=prefix,
                file_path=validation_data_path,
                storage_options=storage_options,
                data_type="validation",
            )
        if train_data_path is None and validation_data_path is None:
            data_type = kwargs.get("data_type", "training")
            if data_type not in ("training", "validation"):
                InvalidDataType(
                    "`data_type` is not supported. Choose 'training' or 'validation'."
                )
            self._save_data_path(prefix, data_type=data_type)

    def _populate_metadata_taxonomy(self, model=None, use_case_type=None):
        """Extract and populate the taxonomy metadata from the model.

        Parameters
        ----------
        model: [sklearn, xgboost, lightgbm, automl, keras]
            The model object.
        use_case_type: str
            The use case type of the model.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError: When model not provided.
        """
        if use_case_type and use_case_type not in UseCaseType:
            raise ValueError(
                f"Invalid value of `UseCaseType`. Choose from {UseCaseType.values()}."
            )

        self.metadata_taxonomy[MetadataTaxonomyKeys.USE_CASE_TYPE].value = use_case_type
        if model is not None:
            map = ModelInfoExtractorFactory.extract_info(model)
            if map is not None:
                self.metadata_taxonomy._populate_from_map(map)
            if (
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].size()
                > METADATA_SIZE_LIMIT
            ):
                logger.warn(
                    f"The model hyperparameters are larger than `{METADATA_SIZE_LIMIT}` "
                    "bytes and cannot be stored as model catalog metadata. It will be saved to "
                    f"{self.artifact_dir}/hyperparameters.json and removed from the metadata."
                )

                self.metadata_taxonomy[
                    MetadataTaxonomyKeys.HYPERPARAMETERS
                ].to_json_file(self.artifact_dir)
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].update(
                    value=None
                )

    def _populate_metadata_custom(self):
        """Extracts custom metadata from the model artifact.

        Returns
        -------
        None
            Nothing
        """
        model_metadata_items = []

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT,
                value=self.conda_env if hasattr(self, "conda_env") else None,
                description="The conda environment where the model was trained.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        try:
            env_type = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_TYPE._value
            )
        except:
            env_type = None
        try:
            slug_name = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_SLUG._value
            )
        except:
            slug_name = None
        try:
            env_path = (
                self._runtime_info.MODEL_DEPLOYMENT.INFERENCE_CONDA_ENV.INFERENCE_ENV_PATH._value
            )
        except:
            env_path = None

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.ENVIRONMENT_TYPE,
                value=env_type,
                description="The environment type, must be a 'published' or 'data_science'.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.SLUG_NAME,
                value=slug_name,
                description="The slug name of the training conda environment.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT_PATH,
                value=env_path,
                description="The oci path of the training conda environment.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_ARTIFACTS,
                value=textwrap.shorten(
                    ", ".join(self._get_files()), 255, placeholder="..."
                ),
                description="A list of files located in the model artifacts folder.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT,
                value=self._serialization_format,
                description="The model serialization format.",
                category=MetadataCustomCategory.TRAINING_PROFILE,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CLIENT_LIBRARY,
                value="ADS",
                description="",
                category=MetadataCustomCategory.OTHER,
            )
        )
        self.metadata_custom._add_many(model_metadata_items, replace=True)

    def populate_metadata(self, model=None, use_case_type=None):
        """Extracts and populate taxonomy metadata from given model.

        Parameters
        ----------
        model: [sklearn, xgboost, lightgbm, automl, keras]
            The model object.

        use_case_type:
            The use case type of the model.
        model: (Any, optional). Defaults to None.
            This is an optional model object which is only used to extract taxonomy metadata.
            Supported models: keras, lightgbm, pytorch, sklearn, tensorflow, and xgboost.
            If the model is not under supported frameworks, then extracting taxonomy metadata will be skipped.
        use_case_type: (str, optional). Default to None.
            The use case type of the model.

        Returns
        -------
        None
            Nothing.
        """
        if model is None and self.metadata_taxonomy["Algorithm"].value is None:
            logger.info(
                "To auto-extract taxonomy metadata the model must be provided. Supported models: automl, keras, lightgbm, pytorch, sklearn, tensorflow, and xgboost."
            )
        if use_case_type is None:
            use_case_type = self.metadata_taxonomy[
                MetadataTaxonomyKeys.USE_CASE_TYPE
            ].value
        self._populate_metadata_taxonomy(model, use_case_type)
        self._populate_metadata_custom()

    def _save_from_memory(
        self,
        data,
        prefix,
        data_file_name,
        storage_options,
        data_type="training",
        **kwargs,
    ):
        """
        Save the data to Object Storage.
        """
        oci_storage_path = os.path.join(prefix, data_file_name)
        if isinstance(data, np.ndarray) or isinstance(data, list):
            data = pd.DataFrame(data)
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        elif isinstance(data, ADSData):
            data = pd.concat([data.X, data.y], axis=1)
            data.to_csv(oci_storage_path, storage_options=storage_options, **kwargs)
        else:
            raise NotImplementedError(
                f"`{type(data)}` is not supported. Use a Pandas DataFrame."
            )

        self._save_data_path(oci_storage_path, data_type)
        self._save_data_shape(data, data_type)

    def _save_from_local_file(
        self, prefix, file_path, storage_options, data_type="training"
    ):
        """Save local file to Object Storage."""
        file_path = os.path.expanduser(file_path)
        import glob

        if len(glob.glob(file_path)) == 0:
            raise FileExistsError(f"No files were found in `{file_path}`.")
        oci_storage_paths = []
        with fsspec.open_files(file_path, mode="r") as fhs:
            for fh in fhs:
                oci_storage_path = os.path.join(prefix, os.path.basename(fh.name))
                with fsspec.open(
                    oci_storage_path,
                    mode="w",
                    **(storage_options),
                ) as f:
                    f.write(fh.read())
                oci_storage_paths.append(oci_storage_path)
                self._save_file_size(
                    os.path.join(os.path.dirname(file_path), os.path.basename(fh.name)),
                    data_type,
                )
        self._save_data_path(",  ".join(oci_storage_paths), data_type)

    def _save_data_path(self, oci_storage_path, data_type):

        key = (
            MetadataCustomKeys.TRAINING_DATASET
            if data_type == "training"
            else MetadataCustomKeys.VALIDATION_DATASET
        )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=key,
                value=oci_storage_path,
                description=f"The path to where the {data_type} dataset is stored on Object Storage.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def _save_data_shape(self, data, data_type):
        key = (
            MetadataCustomKeys.TRAINING_DATASET_SIZE
            if data_type == "training"
            else MetadataCustomKeys.VALIDATION_DATASET_SIZE
        )
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.TRAINING_DATASET_SIZE
                if data_type == "training"
                else MetadataCustomKeys.VALIDATION_DATASET_SIZE,
                value=str(data.shape),
                description=f"The size of the {data_type} dataset in bytes.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def _save_file_size(self, file_path, data_type):
        self.metadata_custom._add(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.TRAINING_DATASET_SIZE
                if data_type == "training"
                else MetadataCustomKeys.VALIDATION_DATASET_SIZE,
                value=str(os.stat(file_path).st_size) + " bytes",
                description=f"The {data_type} dataset size in bytes.",
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            ),
            replace=True,
        )

    def _prepare_data_for_schema(
        self,
        X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
    ):
        """
        Any Framework-specific work before generic schema generation.
        """
        return X_sample, y_sample

    def populate_schema(
        self,
        data_sample: ADSData = None,
        X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        max_col_num: int = DATA_SCHEMA_MAX_COL_NUM,
    ):
        """
        Populate the input and output schema.
        If the schema exceeds the limit of 32kb, save as json files to the artifact directory.

        Parameters
        ----------
        data_sample: ADSData
            A sample of the data that will be used to generate input_schema and output_schema.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of input data that will be used to generate input schema.
        y_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of output data that will be used to generate output schema.
        max_col_num: (int, optional). Defaults to utils.DATA_SCHEMA_MAX_COL_NUM.
            The maximum column size of the data that allows to auto generate schema.
        """
        if data_sample is not None:
            assert isinstance(
                data_sample, ADSData
            ), "`data_sample` expects data of ADSData type. \
            Pass in to `X_sample` and `y_sample` for other data types."
            X_sample = data_sample.X
            y_sample = data_sample.y
        X_sample, y_sample = self._prepare_data_for_schema(X_sample, y_sample)
        self.schema_input = self._populate_schema(
            X_sample,
            schema_file_name=INPUT_SCHEMA_FILE_NAME,
            max_col_num=max_col_num,
        )
        self.schema_output = self._populate_schema(
            y_sample,
            schema_file_name=OUTPUT_SCHEMA_FILE_NAME,
            max_col_num=max_col_num,
        )

    def _populate_schema(
        self,
        data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame],
        schema_file_name: str,
        max_col_num: int,
    ):
        """
        Populate schema and if the schema exceeds the limit of 32kb, save as a json file to artifact_dir.

        Parameters
        ----------
        data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of input data that will be used to generate input schema.
        schema_file_name: str
            schema file name to be saved as.
        max_col_num : int
            The maximum column size of the data that allows to auto generate schema.

        Returns
        -------
        Schema
            The schema.
        """
        result = None

        try:
            if data is not None:
                data = utils.to_dataframe(data)
                schema = data.ads.model_schema(max_col_num=max_col_num)
                schema.to_json_file(os.path.join(self.artifact_dir, schema_file_name))
                if self._validate_schema_size(schema, schema_file_name):
                    result = schema
        except DataSizeTooWide:
            logger.warning(
                f"The data has too many columns and "
                f"the maximum allowable number of columns is `{max_col_num}`. "
                "The schema was not auto generated. Increase allowable number of columns."
            )

        return result or Schema()

    def _validate_schema_size(self, schema, schema_file_name):
        result = False
        try:
            result = schema.validate_size()
        except SchemaSizeTooLarge:
            logger.warn(
                f"The {schema_file_name.replace('.json', '')} is larger than "
                f"`{METADATA_SIZE_LIMIT}` bytes and cannot be stored as model catalog metadata."
                f"It will be saved to {self.artifact_dir}/{schema_file_name}."
            )

        return result

    def introspect(self) -> pd.DataFrame:
        """Runs model introspection.

        Returns
        -------
        pd.DataFrame
           The introspection result in a dataframe format.
        """
        return self._introspect()

    @classmethod
    def from_model_catalog(
        cls,
        model_id: str,
        artifact_dir: str,
        model_file_name: Optional[str] = "model.onnx",
        auth: Optional[Dict] = None,
        force_overwrite: Optional[bool] = False,
        install_libs: Optional[bool] = False,
        conflict_strategy=ConflictStrategy.IGNORE,
        bucket_uri: Optional[str] = None,
        remove_existing_artifact: Optional[bool] = True,
        **kwargs,
    ) -> "ModelArtifact":
        """Download model artifacts from the model catalog to the target artifact directory.

        Parameters
        ----------
        model_id: str
            The model OCID.
        artifact_dir: str
            The artifact directory to store the files needed for deployment.
            Will be created if not exists.
        model_file_name: (str, optional). Defaults to "model.onnx".
            The name of the serialized model.
        auth: (Dict, optional). Defaults to None.
            Default authetication is set using the `ads.set_auth()` method.
            Use the `ads.common.auth.api_keys()` or `ads.common.auth.resource_principal()` to create appropriate
            authentication signer and kwargs required to instantiate a IdentityClient object.
        force_overwrite: (bool, optional). Defaults to False.
            Overwrite existing files.
        install_libs: bool, default: False
            Install the libraries specified in ds-requirements.txt.
        conflict_strategy: ConflictStrategy, default: IGNORE
           Determines how to handle version conflicts between the current environment and requirements of
           model artifact.
           Valid values: "IGNORE", "UPDATE" or ConflictStrategy.
           IGNORE: Use the installed version in  case of conflict
           UPDATE: Force update dependency to the version required by model artifact in case of conflict
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts with
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.
        kwargs:
            compartment_id: (str, optional)
                Compartment OCID. If not specified, the value will be taken from the environment variables.
            timeout: (int, optional). Defaults to 10 seconds.
                The connection timeout in seconds for the client.

        Returns
        -------
        ModelArtifact
            An instance of ModelArtifact class.
        """
        from ads.catalog.model import ModelCatalog

        auth = auth or authutil.default_signer()
        artifact_dir = os.path.abspath(os.path.expanduser(artifact_dir))

        model_catalog = ModelCatalog(
            compartment_id=kwargs.pop("compartment_id", _COMPARTMENT_OCID),
            ds_client_auth=auth,
            identity_client_auth=auth,
            timeout=kwargs.pop("timeout", None),
        )

        model_catalog._download_artifact(
            model_id=model_id,
            target_dir=artifact_dir,
            force_overwrite=force_overwrite,
            bucket_uri=bucket_uri,
            remove_existing_artifact=remove_existing_artifact,
        )
        oci_model = model_catalog.get_model(model_id)

        result_artifact = cls(
            artifact_dir=artifact_dir,
            conflict_strategy=conflict_strategy,
            install_libs=install_libs,
            reload=False,
            model_file_name=model_file_name,
        )

        result_artifact.metadata_custom = oci_model.metadata_custom
        result_artifact.metadata_taxonomy = oci_model.metadata_taxonomy
        result_artifact.schema_input = oci_model.schema_input
        result_artifact.schema_output = oci_model.schema_output

        if not install_libs:
            logger.warning(
                "Libraries in `ds-requirements.txt` were not installed. "
                "Use `install_requirements()` to install the required dependencies."
            )

        return result_artifact


class VersionConflictWarning(object):
    def __init__(self, version_conflicts):
        self.version_conflicts = version_conflicts

    def __str__(self):
        msg = "WARNING: Version conflicts found:"
        if len(self.version_conflicts) > 0:
            for lib in self.version_conflicts:
                msg += "\nInstalled: %s, Required: %s" % (
                    lib,
                    self.version_conflicts[lib],
                )
        return msg


def pip_install(package, options="-U"):
    package = re.sub(r"<|>", "=", package.split(",")[0])
    for output in execute(["pip", "install", options, package]):
        print(output, end="")


def execute(cmd):
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import itertools
from typing import Dict, List

from ads.common.decorator.runtime_dependency import (
    OptionalDependency,
    runtime_dependency,
)
from ads.common.extended_enum import ExtendedEnumMeta
from ads.pipeline.visualizer.base import (
    COMPLETE_STATUS,
    FAIL_STATUS,
    STATUS_COLOR_MAP,
    STATUS_TEXT_MAP,
    STEP_KIND_MAP,
    GraphOrientation,
    PipelineRenderer,
    RendererItem,
    RendererItemStatus,
    StepStatus,
    _strfdelta,
)

GRAPH_BOX_COLOR = "#DEDEDE"


class RenderTo(str, metaclass=ExtendedEnumMeta):
    SVG = "svg"
    JPEG = "jpeg"
    PNG = "png"


STEP_WITH_STATUS_LABEL_TEMPLATE = """<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="2">
        <TR><TD><FONT COLOR="white" POINT-SIZE="14.0" FACE="Helvetica,Arial,sans-serif">{step_name}</FONT></TD></TR>
        <TR BORDER="1"><TD><FONT COLOR="white" POINT-SIZE="11.0" FACE="Helvetica,Arial,sans-serif">{step_kind}</FONT></TD></TR>
        <TR><TD><FONT COLOR="white" POINT-SIZE="11.0"  FACE="Courier New">{status_name}</FONT></TD></TR>
        <TR><TD><FONT COLOR="white" POINT-SIZE="11.0" FACE="Courier New">{duration}&nbsp;</FONT></TD></TR>
        </TABLE>>"""

STEP_LABEL_TEMPLATE = """<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="2">
    <TR><TD><FONT COLOR="white" POINT-SIZE="14.0" FACE="Helvetica,Arial,sans-serif">{step_name}</FONT></TD></TR>
    <TR><TD><FONT COLOR="white" POINT-SIZE="11.0" FACE="Helvetica,Arial,sans-serif">{step_kind}</FONT></TD></TR>
    </TABLE>>"""


class PipelineGraphRenderer(PipelineRenderer):
    def __init__(self, show_status: bool = False):
        """Initialize a PipelineGraphRenderer class.

        Parameters
        ----------
        show_status : bool, defaults to False.
            Whether to display status for steps.

        Returns
        -------
        None
            Nothing.
        """
        super().__init__()
        self.show_status = show_status

    def _add_final_step(self):
        """
        Add final step when rendering pipeline step with status in graph.

        Returns
        -------
        None
        """
        final_step_name = "Done"

        all_steps_set = set([step.key for step in self.steps])
        steps_deps_set = set(itertools.chain(*self.deps.values()))
        final_nodes = list(all_steps_set - steps_deps_set)
        final_step_status = StepStatus.WAITING
        if all(
            single_step_status.lifecycle_state == StepStatus.SUCCEEDED
            for single_step_status in self.step_status.values()
        ):
            final_step_status = StepStatus.SUCCEEDED
        elif any(
            single_step_status.lifecycle_state in FAIL_STATUS
            for single_step_status in self.step_status.values()
        ) and all(
            single_step_status.lifecycle_state in COMPLETE_STATUS
            for single_step_status in self.step_status.values()
        ):
            final_step_status = StepStatus.FAILED

        self.graph.node(
            name=final_step_name,
            label=f'<<FONT POINT-SIZE="16.0" color="white" FACE="Helvetica,Arial,sans-serif"><B>{final_step_name}</B></FONT>>',
            shape="tripleoctagon",
            style="filled, rounded",
            fontsize="18.0",
            color=GRAPH_BOX_COLOR,
            fillcolor=STATUS_COLOR_MAP[final_step_status],
        )

        for node in final_nodes:
            self.graph.edge(node, final_step_name)

    @runtime_dependency(module="graphviz", install_from=OptionalDependency.VIZ)
    def _generate_graph(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[RendererItem]] = None,
        step_status: Dict[str, RendererItemStatus] = None,
        rankdir: str = GraphOrientation.TOP_BOTTOM,
        **kwargs,
    ):
        """
        Generates Pipeline graph.

        Parameters
        ----------
        steps: List[RendererItem]
            A list of RendererItem objects.
        deps: Dict[str, List[RendererItem]]
            A dictionary mapping the key of a RendererItem to a list of
            RendererItem that this step depends on.
        step_status: Dict[str, RendererItemStatus], defaults to None.
            A dictionary mapping the key of a RendererItem to its current
            status.
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.

        Returns
        -------
        None
        """
        from graphviz import Digraph

        self.steps = steps
        self.deps = deps
        self.step_status = step_status

        if self.show_status and not step_status:
            raise ValueError(
                "`step_status` must be provided to render step status in graph."
            )

        self.graph = Digraph(graph_attr={"rankdir": rankdir})
        self.graph.attr("node", shape="box")

        for step in steps:
            if self.show_status and step_status:
                label = STEP_WITH_STATUS_LABEL_TEMPLATE.format(
                    step_name=step.name,
                    status_name=STATUS_TEXT_MAP[step_status[step.key].lifecycle_state],
                    step_kind=STEP_KIND_MAP[step.kind],
                    duration=_strfdelta(step_status[step.key].duration),
                )
                step_fillcolor = STATUS_COLOR_MAP[step_status[step.key].lifecycle_state]

            else:
                label = STEP_LABEL_TEMPLATE.format(
                    step_name=step.name,
                    step_kind=STEP_KIND_MAP[step.kind],
                )
                step_fillcolor = STATUS_COLOR_MAP[StepStatus.WAITING]

            self.graph.node(
                name=step.key,
                label=label,
                shape=None,
                style="filled, rounded",
                fontsize="11",
                color=GRAPH_BOX_COLOR,
                fillcolor=step_fillcolor,
            )
            for dep in deps[step.key]:
                self.graph.edge(dep.key, step.key)

        if self.show_status and step_status:
            self._add_final_step()

    def render(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[RendererItem]] = None,
        step_status: Dict[str, RendererItemStatus] = None,
        rankdir: str = GraphOrientation.TOP_BOTTOM,
        **kwargs,
    ):
        """
        Renders Pipeline graph.

        Parameters
        ----------
        steps: List[RendererItem]
            A list of RendererItem objects.
        deps: Dict[str, List[RendererItem]]
            A dictionary mapping the key of a RendererItem to a list of
            RendererItem that this step depends on.
        step_status: Dict[str, RendererItemStatus], defaults to None.
            A dictionary mapping the key of a RendererItem to its current
            status.
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.

        Returns
        -------
        None
        """
        self._generate_graph(
            steps=steps,
            deps=deps,
            step_status=step_status,
            rankdir=rankdir,
        )
        try:
            from IPython.core.display import display

            display(self.graph)
        except:
            pass

    def save_to(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[RendererItem]] = None,
        step_status: Dict[str, RendererItemStatus] = None,
        rankdir: str = GraphOrientation.TOP_BOTTOM,
        uri: str = None,
        format: str = RenderTo.SVG,
        **kwargs,
    ) -> str:
        """
        Renders pipeline as graph in selected format.

        steps: List[RendererItem]
            A list of RendererItem objects.
        deps: Dict[str, List[RendererItem]]
            A dictionary mapping the key of a RendererItem to a list of
            RendererItem that this step depends on.
        step_status: Dict[str, RendererItemStatus], defaults to None.
            A dictionary mapping the key of a RendererItem to its current
            status.
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.
        uri: (string, optional). Defaults to None.
            URI location to save the SVG string.
        format: (str, optional). Defaults to "svg".
            The format to save the graph.
            Supported formats: "svg", "html".

        Returns
        -------
        str
            Graph in selected format.
        """
        if format not in RenderTo:
            raise ValueError(
                f"Unsupported format: `{format}`. Supported formats: {RenderTo.values()}"
            )

        self._generate_graph(
            steps=steps,
            deps=deps,
            step_status=step_status,
            rankdir=rankdir,
        )
        result = self.graph.pipe(format=format)
        if uri:
            self._write_to_file(content=result, uri=uri, mode="b", **kwargs)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC

from ads.hpo.distributions import *
from ads.hpo.utils import _update_space_name

from sklearn.ensemble import (
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LogisticRegression,
    Ridge,
    RidgeClassifier,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.svm import SVC, SVR, LinearSVC, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


class ModelSearchSpace(ABC):
    """Defines an abstract base class for setting the search space and strategy used
    during hyperparameter optimization
    """

    def __init__(self, strategy):
        self.strategy = strategy
        self.space = {}
        self.step_name = ""
        super().__init__()

    def suggest_space(self, **kwargs):
        pass


class RidgeSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(RidgeSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {"alpha": LogUniformDistribution(10**-4, 10**-1)}
        if self.strategy != "perfunctory":
            space.update(
                {
                    "alpha": LogUniformDistribution(10**-5, 10**1),
                    "fit_intercept": CategoricalDistribution([True, False]),
                    "normalize": CategoricalDistribution([True, False]),
                }
            )
        return _update_space_name(space, **kwargs)


class LassoSearchSpace(RidgeSearchSpace):
    pass


class RidgeClassifierSearchSpace(RidgeSearchSpace):
    pass


class ElasticNetSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(ElasticNetSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {
            "alpha": LogUniformDistribution(10**-4, 10**-1),
            "l1_ratio": UniformDistribution(0, 1),
        }
        if self.strategy != "perfunctory":
            space.update(
                {
                    "alpha": LogUniformDistribution(10**-5, 10),
                    "fit_intercept": CategoricalDistribution([True, False]),
                    "l1_ratio": UniformDistribution(0, 1),
                    "normalize": CategoricalDistribution([True, False]),
                }
            )
        return _update_space_name(space, **kwargs)


class LogisticRegressionSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(LogisticRegressionSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {
            "C": LogUniformDistribution(10**-4, 10**-1),
            "dual": CategoricalDistribution([False]),
            "penalty": CategoricalDistribution(["l1", "l2"]),
            "solver": CategoricalDistribution(["saga"]),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "C": LogUniformDistribution(10**-5, 10),
                    "l1_ratio": UniformDistribution(0, 1),
                    "penalty": CategoricalDistribution(["elasticnet"]),
                }
            )
        return _update_space_name(space, **kwargs)


class SGDClassifierSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(SGDClassifierSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {
            "alpha": LogUniformDistribution(10**-4, 10**-1),
            "penalty": CategoricalDistribution(["l1", "l2", "none"]),
        }
        if self.strategy != "perfunctory":
            space.update(
                {
                    "alpha": LogUniformDistribution(10**-5, 10**1),
                    "l1_ratio": UniformDistribution(0, 1),
                    "penalty": CategoricalDistribution(["elasticnet"]),
                }
            )
        return _update_space_name(space, **kwargs)


class SGDRegressorSearchSpace(SGDClassifierSearchSpace):
    def __init__(self, strategy):
        super(SGDRegressorSearchSpace, self).__init__(strategy)


class SVCSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(SVCSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "C": LogUniformDistribution(10**-4, 10**-1),
            "max_iter": CategoricalDistribution([1000]),
        }
        if self.strategy != "perfunctory":
            space.update(
                {
                    "C": LogUniformDistribution(10**-5, 5),
                    "gamma": CategoricalDistribution(["scale", "auto"]),
                    "kernel": CategoricalDistribution(
                        ["linear", "poly", "rbf", "sigmoid"]
                    ),
                    "max_iter": CategoricalDistribution([5000]),
                }
            )
        return _update_space_name(space, **kwargs)


class SVRSearchSpace(SVCSearchSpace):
    pass


class LinearSVCSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(LinearSVCSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "C": LogUniformDistribution(10**-4, 10**-1),
            "dual": CategoricalDistribution([False]),
        }
        if self.strategy != "perfunctory":
            space.update(
                {
                    "C": LogUniformDistribution(10**-5, 5),
                    "class_weight": CategoricalDistribution(
                        ["balanced", None]
                    ),  # max_iter defaults to 1000
                    "fit_intercept": CategoricalDistribution([True, False]),
                    "loss": CategoricalDistribution(["squared_hinge"]),
                    "penalty": CategoricalDistribution(["l1"]),
                }
            )
        return _update_space_name(space, **kwargs)


class LinearSVRSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(LinearSVRSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {"C": LogUniformDistribution(10**-4, 10**-1)}

        if self.strategy != "perfunctory":
            space.update(
                {
                    "C": LogUniformDistribution(10**-5, 10**1),
                    "dual": CategoricalDistribution([False]),
                    "fit_intercept": CategoricalDistribution([True, False]),
                    "loss": CategoricalDistribution(["squared_epsilon_insensitive"]),
                }
            )
        return _update_space_name(space, **kwargs)


class DecisionTreeClassifierSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(DecisionTreeClassifierSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "max_depth": IntUniformDistribution(1, 5),
            "min_impurity_decrease": UniformDistribution(0, 0.05),
            "min_samples_split": IntUniformDistribution(2, 500),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "criterion": CategoricalDistribution(["gini", "entropy"]),
                    "max_depth": IntUniformDistribution(1, 10),
                    "min_samples_leaf": IntUniformDistribution(2, 500),
                }
            )  # max_iter defaults to 1000

        return _update_space_name(space, **kwargs)


class DecisionTreeRegressorSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(DecisionTreeRegressorSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "max_depth": IntUniformDistribution(1, 5),
            "min_impurity_decrease": UniformDistribution(0, 0.05),
            "min_samples_split": IntUniformDistribution(2, 500),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "criterion": CategoricalDistribution(
                        ["mse", "friedman_mse", "mae"]
                    ),
                    "min_samples_leaf": IntUniformDistribution(2, 500),
                }
            )  # max_iter defaults to 1000

        return _update_space_name(space, **kwargs)


class XGBClassifierSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(XGBClassifierSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {
            "colsample_bytree": UniformDistribution(0.6, 0.8),
            "learning_rate": LogUniformDistribution(0.3, 0.4),
            "max_depth": IntUniformDistribution(1, 5),
            "n_estimators": IntUniformDistribution(50, 250),
            "subsample": UniformDistribution(0.5, 1),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "colsample_bytree": UniformDistribution(0.3, 0.7),
                    "gamma": UniformDistribution(0, 10),
                    "learning_rate": LogUniformDistribution(0.001, 0.6),
                    "max_depth": IntUniformDistribution(1, 10),
                    "min_child_weight": IntUniformDistribution(0, 20),
                    "n_estimators": IntUniformDistribution(50, 500),
                    # 'scale_pos_weight': LogUniformDistribution(10 ** -5, 1),
                    "subsample": UniformDistribution(0.25, 1),
                    "reg_alpha": LogUniformDistribution(10**-5, 1),
                    "reg_lambda": LogUniformDistribution(10**-5, 1),
                }
            )
        return _update_space_name(space, **kwargs)


class XGBRegressorSearchSpace(XGBClassifierSearchSpace):
    pass


class LGBMClassifierSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(LGBMClassifierSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):
        space = {
            "colsample_bytree": UniformDistribution(0.3, 0.7),
            "learning_rate": LogUniformDistribution(0.3, 0.4),
            "max_depth": IntUniformDistribution(1, 5),
            "n_estimators": IntUniformDistribution(50, 250),
            "subsample": UniformDistribution(0.5, 1),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "boosting_type": CategoricalDistribution(["dart"]),
                    "learning_rate": LogUniformDistribution(0.001, 0.6),
                    "max_depth": IntUniformDistribution(1, 10),
                    "min_child_weight": IntUniformDistribution(0, 20),
                    "n_estimators": IntUniformDistribution(50, 500),
                    "num_leaves": IntLogUniformDistribution(7, 40),
                    "reg_alpha": LogUniformDistribution(10**-5, 1),
                    "reg_lambda": LogUniformDistribution(10**-5, 1),
                }
            )

        return _update_space_name(space, **kwargs)


class LGBMRegressorSearchSpace(LGBMClassifierSearchSpace):
    pass


class ExtraTreesClassifierSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(ExtraTreesClassifierSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "n_estimators": IntUniformDistribution(50, 250),
            "max_depth": IntUniformDistribution(1, 5),
            "max_features": CategoricalDistribution(["sqrt", "log2"]),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "max_depth": IntUniformDistribution(1, 10),
                    "min_impurity_decrease": UniformDistribution(0.0, 0.05),
                    "min_samples_split": IntUniformDistribution(2, 500),
                    "min_samples_leaf": IntUniformDistribution(5, 25),
                    "min_weight_fraction_leaf": UniformDistribution(0.0, 0.5),
                    "n_estimators": IntUniformDistribution(50, 500),
                }
            )

        return _update_space_name(space, **kwargs)


class ExtraTreesRegressorSearchSpace(ExtraTreesClassifierSearchSpace):
    pass


class RandomForestClassifierSearchSpace(ExtraTreesClassifierSearchSpace):
    pass


class RandomForestRegressorSearchSpace(ExtraTreesClassifierSearchSpace):
    pass


class GradientBoostingRegressorSearchSpace(ModelSearchSpace):
    def __init__(self, strategy):
        super(GradientBoostingRegressorSearchSpace, self).__init__(strategy)

    def suggest_space(self, **kwargs):

        space = {
            "max_depth": IntUniformDistribution(1, 5),
            "max_features": CategoricalDistribution(["sqrt", "log2"]),
            "n_estimators": IntUniformDistribution(50, 250),
        }

        if self.strategy != "perfunctory":
            space.update(
                {
                    "learning_rate": LogUniformDistribution(0.001, 0.6),
                    "max_depth": IntUniformDistribution(1, 10),
                    "min_samples_leaf": IntUniformDistribution(5, 25),
                    "min_samples_split": IntUniformDistribution(2, 500),
                    "n_estimators": IntUniformDistribution(50, 500),
                    "subsample": UniformDistribution(0.5, 1),
                }
            )

        return _update_space_name(space, **kwargs)


class GradientBoostingClassifierSearchSpace(GradientBoostingRegressorSearchSpace):
    pass


def get_model2searchspace():
    model2searchspace = {
        Ridge: RidgeSearchSpace,
        RidgeClassifier: RidgeClassifierSearchSpace,
        Lasso: LassoSearchSpace,
        ElasticNet: ElasticNetSearchSpace,
        LogisticRegression: LogisticRegressionSearchSpace,
        SVC: SVCSearchSpace,
        SVR: SVRSearchSpace,
        LinearSVC: LinearSVCSearchSpace,
        LinearSVR: LinearSVRSearchSpace,
        DecisionTreeClassifier: DecisionTreeClassifierSearchSpace,
        DecisionTreeRegressor: DecisionTreeRegressorSearchSpace,
        RandomForestClassifier: RandomForestClassifierSearchSpace,
        RandomForestRegressor: RandomForestRegressorSearchSpace,
        GradientBoostingClassifier: GradientBoostingClassifierSearchSpace,
        GradientBoostingRegressor: GradientBoostingRegressorSearchSpace,
        ExtraTreesClassifier: ExtraTreesClassifierSearchSpace,
        ExtraTreesRegressor: ExtraTreesRegressorSearchSpace,
        SGDClassifier: SGDClassifierSearchSpace,
        SGDRegressor: SGDRegressorSearchSpace,
    }
    try:
        from xgboost import XGBClassifier, XGBRegressor

        model2searchspace[XGBClassifier] = XGBClassifierSearchSpace
        model2searchspace[XGBRegressor] = XGBRegressorSearchSpace
    except:
        pass
    try:
        from lightgbm import LGBMClassifier, LGBMRegressor

        model2searchspace[LGBMClassifier] = LGBMClassifierSearchSpace
        model2searchspace[LGBMRegressor] = LGBMRegressorSearchSpace
    except:
        pass
    return model2searchspace


model_list = list(get_model2searchspace().keys())



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import re

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import ZipcodeTypedFeature


class ZipCodeDetector(AbstractTypeDiscoveryDetector):
    def _is_zip_code(self, values):
        return all(
            [re.match("^[0-9]{5}(?:-[0-9]{4})?$", str(v)) for v in values.head(10)]
        )

    def discover(self, name, series):
        candidates = series.loc[~series.isnull()]
        if (
            self._is_zip_code(candidates.head(1000))
            if candidates.dtype == "object"
            else self._is_zip_code(candidates.astype("object"))
        ):
            logger.debug(
                "type discovery on column [{}]/[{}] found to be a zipcode".format(
                    name, series.dtype
                )
            )
            return ZipcodeTypedFeature.build(name, series)

        return False


if __name__ == "__main__":
    dd = ZipCodeDetector()
    print(
        dd.discover(
            "zip", pd.Series([None, "00501", "94065", "94065-1107", "90210", None])
        )
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
    This boilerplate is based on a n sklearn model serialized with cloudpickle.
"""
import json
import os


model_name = "model.pkl"


def load_model(model_file_name=model_name):
    """
    Loads model from the serialized format
    Returns
    -------
    model:  a model instance on which predict API can be invoked
    """
    model_dir = os.path.dirname(os.path.realpath(__file__))
    contents = os.listdir(model_dir)
    if model_file_name in contents:
        from cloudpickle import cloudpickle

        with open(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), model_file_name),
            "rb",
        ) as file:
            return cloudpickle.load(file)
    else:
        raise Exception(
            "{0} is not found in model directory {1}".format(model_file_name, model_dir)
        )


def predict(data, model=load_model()):
    """
    Returns prediction given the model and data to predict
    Parameters
    ----------
    model: Model instance returned by load_model API
    data: Data format as expected by the predict API of the core estimator. For eg. in case of sckit models it could be numpy array/List of list/Panda DataFrame
    Returns
    -------
    predictions: Output from scoring server
        Format: {'prediction':output from model.predict method}
    """
    from pandas import read_json, DataFrame
    from io import StringIO

    data = (
        read_json(StringIO(data))
        if isinstance(data, str)
        else DataFrame.from_dict(data)
    )
    pred = model.predict(data).tolist()
    return {"prediction": pred}



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import time
from abc import abstractmethod

from oci._vendor import six
from tqdm.auto import tqdm


class ProgressBar(object):
    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def update(self, description):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TqdmProgressBar(ProgressBar):
    def __enter__(self):
        return self

    def __init__(self, max_progress=100, description="Running", verbose=False):
        self.max_progress = max_progress
        self.progress_bar = tqdm(
            total=max_progress, desc="loop1", leave=False, mininterval=0
        )

        self.progress_bar.set_description(description)
        self.verbose = verbose
        if self.verbose:
            self.start_time = time.time()
            self.description = description

    def update(self, description=None):
        """
        Updates the progress bar
        """
        if self.verbose and description is not None:
            print(
                "%s:%ss"
                % (
                    self.description,
                    str.format("{0:.3f}", time.time() - self.start_time),
                )
            )
            self.description = (
                description if description is not None else self.description
            )
            self.start_time = time.time()
        self.progress_bar.update(1)
        if description is not None:
            self.progress_bar.set_description(description, refresh=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_tb is None:
            self.progress_bar.set_description("Done")
            self.progress_bar.close()
            return
        six.reraise(exc_type, exc_val, exc_tb)


class DummyProgressBar(ProgressBar):
    def __enter__(self):
        return self

    def __init__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        """
        Updates the progress bar
        """
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.model.generic_model import GenericModel, ModelState
from ads.model.datascience_model import DataScienceModel
from ads.model.model_properties import ModelProperties
from ads.model.framework.automl_model import AutoMLModel
from ads.model.framework.lightgbm_model import LightGBMModel
from ads.model.framework.pytorch_model import PyTorchModel
from ads.model.framework.sklearn_model import SklearnModel
from ads.model.framework.tensorflow_model import TensorFlowModel
from ads.model.framework.xgboost_model import XGBoostModel
from ads.model.framework.spark_model import SparkPipelineModel
from ads.model.framework.huggingface_model import HuggingFacePipelineModel

from ads.model.deployment.model_deployer import ModelDeployer
from ads.model.deployment.model_deployment import ModelDeployment
from ads.model.deployment.model_deployment_properties import ModelDeploymentProperties

from ads.model.serde.common import SERDE
from ads.model.serde.model_input import ModelInputSerializer

from ads.model.model_version_set import ModelVersionSet, experiment
from ads.model.service.oci_datascience_model_version_set import (
    ModelVersionSetNotExists,
    ModelVersionSetNotSaved,
)
